# Generated by Django 3.0.5 on 2020-04-14 18:12

import private_storage.fields
import private_storage.storage.files
from django.db import migrations

import openwisp_firmware_upgrader.base.models


class Migration(migrations.Migration):
    dependencies = [
        ('firmware_upgrader', '0002_default_permissions'),
    ]

    operations = [
        migrations.AlterField(
            model_name='firmwareimage',
            name='file',
            field=private_storage.fields.PrivateFileField(
                storage=private_storage.storage.files.PrivateFileSystemStorage(),
                upload_to=openwisp_firmware_upgrader.base.models.get_build_directory,
                verbose_name='File',
            ),
        ),
    ]



import logging

import swapper
from celery import shared_task
from celery.exceptions import SoftTimeLimitExceeded
from django.core.exceptions import ObjectDoesNotExist
from django.utils.translation import gettext_lazy as _

from openwisp_utils.tasks import OpenwispCeleryTask

from . import settings as app_settings
from .exceptions import RecoverableFailure
from .swapper import load_model

logger = logging.getLogger(__name__)


@shared_task(
    bind=True,
    autoretry_for=(RecoverableFailure,),
    soft_time_limit=app_settings.TASK_TIMEOUT,
    **app_settings.RETRY_OPTIONS,
)
def upgrade_firmware(self, operation_id):
    """
    Calls the ``upgrade()`` method of an
    ``UpgradeOperation`` instance in the background
    """
    try:
        operation = load_model('UpgradeOperation').objects.get(pk=operation_id)
        recoverable = self.request.retries < self.max_retries
        operation.upgrade(recoverable=recoverable)
    except SoftTimeLimitExceeded:
        operation.status = 'failed'
        operation.log_line(_('Operation timed out.'))
        logger.warning('SoftTimeLimitExceeded raised in upgrade_firmware task')
    except ObjectDoesNotExist:
        logger.warning(
            f'The UpgradeOperation object with id {operation_id} has been deleted'
        )


@shared_task(bind=True, soft_time_limit=app_settings.TASK_TIMEOUT)
def batch_upgrade_operation(self, batch_id, firmwareless):
    """
    Calls the ``batch_upgrade()`` method of a
    ``Build`` instance in the background
    """
    try:
        batch_operation = load_model('BatchUpgradeOperation').objects.get(pk=batch_id)
        batch_operation.upgrade(firmwareless=firmwareless)
    except SoftTimeLimitExceeded:
        batch_operation.status = 'failed'
        batch_operation.save()
        logger.warning('SoftTimeLimitExceeded raised in batch_upgrade_operation task')
    except ObjectDoesNotExist:
        logger.warning(
            f'The BatchUpgradeOperation object with id {batch_id} has been deleted'
        )


@shared_task(base=OpenwispCeleryTask, bind=True)
def create_device_firmware(self, device_id):
    DeviceFirmware = load_model('DeviceFirmware')
    Device = swapper.load_model('config', 'Device')

    qs = DeviceFirmware.objects.filter(device_id=device_id)
    if qs.exists():
        return

    device = Device.objects.get(pk=device_id)
    DeviceFirmware.create_for_device(device)


@shared_task(base=OpenwispCeleryTask, bind=True)
def create_all_device_firmwares(self, firmware_image_id):
    DeviceFirmware = load_model('DeviceFirmware')
    FirmwareImage = load_model('FirmwareImage')
    Device = swapper.load_model('config', 'Device')

    fw_image = FirmwareImage.objects.select_related('build').get(pk=firmware_image_id)

    queryset = Device.objects.filter(os=fw_image.build.os)
    for device in queryset.iterator():
        DeviceFirmware.create_for_device(device, fw_image)



import uuid

import django.db.models.deletion
import django.utils.timezone
import model_utils.fields
import swapper
from django.conf import settings
from django.db import migrations, models
from swapper import dependency, split

import openwisp_users.mixins

from ..hardware import FIRMWARE_IMAGE_TYPE_CHOICES
from ..swapper import get_model_name


class Migration(migrations.Migration):
    initial = True

    dependencies = [
        ('config', '0015_default_groups_permissions'),
        dependency(*split(settings.AUTH_USER_MODEL), version='0004_default_groups'),
        swapper.dependency('firmware_upgrader', 'Category'),
        swapper.dependency('firmware_upgrader', 'Build'),
        swapper.dependency('firmware_upgrader', 'FirmwareImage'),
        swapper.dependency('firmware_upgrader', 'DeviceFirmware'),
        swapper.dependency('firmware_upgrader', 'BatchUpgradeOperation'),
        swapper.dependency('firmware_upgrader', 'UpgradeOperation'),
    ]

    operations = [
        migrations.CreateModel(
            name='BatchUpgradeOperation',
            fields=[
                (
                    'id',
                    models.UUIDField(
                        default=uuid.uuid4,
                        editable=False,
                        primary_key=True,
                        serialize=False,
                    ),
                ),
                (
                    'created',
                    model_utils.fields.AutoCreatedField(
                        default=django.utils.timezone.now,
                        editable=False,
                        verbose_name='created',
                    ),
                ),
                (
                    'modified',
                    model_utils.fields.AutoLastModifiedField(
                        default=django.utils.timezone.now,
                        editable=False,
                        verbose_name='modified',
                    ),
                ),
                (
                    'status',
                    models.CharField(
                        choices=[
                            ('in-progress', 'in progress'),
                            ('success', 'completed successfully'),
                            ('failed', 'completed with some failures'),
                        ],
                        default='in-progress',
                        max_length=12,
                    ),
                ),
            ],
            options={
                'swappable': swapper.swappable_setting(
                    'firmware_upgrader', 'BatchUpgradeOperation'
                ),
                'verbose_name_plural': 'Mass upgrade operations',
                'verbose_name': 'Mass upgrade operation',
            },
        ),
        migrations.CreateModel(
            name='Build',
            fields=[
                (
                    'id',
                    models.UUIDField(
                        default=uuid.uuid4,
                        editable=False,
                        primary_key=True,
                        serialize=False,
                    ),
                ),
                (
                    'created',
                    model_utils.fields.AutoCreatedField(
                        default=django.utils.timezone.now,
                        editable=False,
                        verbose_name='created',
                    ),
                ),
                (
                    'modified',
                    model_utils.fields.AutoLastModifiedField(
                        default=django.utils.timezone.now,
                        editable=False,
                        verbose_name='modified',
                    ),
                ),
                ('version', models.CharField(db_index=True, max_length=32)),
                (
                    'changelog',
                    models.TextField(
                        blank=True,
                        help_text='descriptive text indicating what has changed since the previous version, if applicable',
                        verbose_name='change log',
                    ),
                ),
            ],
            options={
                'swappable': swapper.swappable_setting('firmware_upgrader', 'Build'),
                'ordering': ('-created',),
                'verbose_name': 'Firmware Build',
                'verbose_name_plural': 'Firmware Builds',
            },
        ),
        migrations.CreateModel(
            name='Category',
            fields=[
                (
                    'id',
                    models.UUIDField(
                        default=uuid.uuid4,
                        editable=False,
                        primary_key=True,
                        serialize=False,
                    ),
                ),
                (
                    'created',
                    model_utils.fields.AutoCreatedField(
                        default=django.utils.timezone.now,
                        editable=False,
                        verbose_name='created',
                    ),
                ),
                (
                    'modified',
                    model_utils.fields.AutoLastModifiedField(
                        default=django.utils.timezone.now,
                        editable=False,
                        verbose_name='modified',
                    ),
                ),
                ('name', models.CharField(db_index=True, max_length=64)),
                ('description', models.TextField(blank=True)),
                (
                    'organization',
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        to=swapper.get_model_name('openwisp_users', 'Organization'),
                        verbose_name='organization',
                    ),
                ),
            ],
            options={
                'swappable': swapper.swappable_setting('firmware_upgrader', 'Category'),
                'verbose_name': 'Firmware Category',
                'verbose_name_plural': 'Firmware Categories',
            },
            bases=(openwisp_users.mixins.ValidateOrgMixin, models.Model),
        ),
        migrations.CreateModel(
            name='DeviceFirmware',
            fields=[
                (
                    'id',
                    models.UUIDField(
                        default=uuid.uuid4,
                        editable=False,
                        primary_key=True,
                        serialize=False,
                    ),
                ),
                (
                    'created',
                    model_utils.fields.AutoCreatedField(
                        default=django.utils.timezone.now,
                        editable=False,
                        verbose_name='created',
                    ),
                ),
                (
                    'modified',
                    model_utils.fields.AutoLastModifiedField(
                        default=django.utils.timezone.now,
                        editable=False,
                        verbose_name='modified',
                    ),
                ),
                ('installed', models.BooleanField(default=False)),
                (
                    'device',
                    models.OneToOneField(
                        on_delete=django.db.models.deletion.CASCADE,
                        to=swapper.get_model_name('config', 'Device'),
                    ),
                ),
            ],
            options={
                'swappable': swapper.swappable_setting(
                    'firmware_upgrader', 'DeviceFirmware'
                ),
                'abstract': False,
                'verbose_name': 'Device Firmware',
            },
        ),
        migrations.CreateModel(
            name='FirmwareImage',
            fields=[
                (
                    'id',
                    models.UUIDField(
                        default=uuid.uuid4,
                        editable=False,
                        primary_key=True,
                        serialize=False,
                    ),
                ),
                (
                    'created',
                    model_utils.fields.AutoCreatedField(
                        default=django.utils.timezone.now,
                        editable=False,
                        verbose_name='created',
                    ),
                ),
                (
                    'modified',
                    model_utils.fields.AutoLastModifiedField(
                        default=django.utils.timezone.now,
                        editable=False,
                        verbose_name='modified',
                    ),
                ),
                ('file', models.FileField(upload_to='')),
                (
                    'type',
                    models.CharField(
                        blank=True,
                        choices=FIRMWARE_IMAGE_TYPE_CHOICES,
                        help_text='firmware image type: model or architecture. Leave blank to attempt determining automatically',
                        max_length=128,
                    ),
                ),
                (
                    'build',
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        to=get_model_name('Build'),
                    ),
                ),
            ],
            options={
                'swappable': swapper.swappable_setting(
                    'firmware_upgrader', 'FirmwareImage'
                ),
                'abstract': False,
                'verbose_name': 'Firmware Image',
                'verbose_name_plural': 'Firmware Images',
            },
        ),
        migrations.CreateModel(
            name='UpgradeOperation',
            fields=[
                (
                    'id',
                    models.UUIDField(
                        default=uuid.uuid4,
                        editable=False,
                        primary_key=True,
                        serialize=False,
                    ),
                ),
                (
                    'created',
                    model_utils.fields.AutoCreatedField(
                        default=django.utils.timezone.now,
                        editable=False,
                        verbose_name='created',
                    ),
                ),
                (
                    'modified',
                    model_utils.fields.AutoLastModifiedField(
                        default=django.utils.timezone.now,
                        editable=False,
                        verbose_name='modified',
                    ),
                ),
                (
                    'status',
                    models.CharField(
                        choices=[
                            ('in-progress', 'in progress'),
                            ('success', 'success'),
                            ('failed', 'failed'),
                            ('aborted', 'aborted'),
                        ],
                        default='in-progress',
                        max_length=12,
                    ),
                ),
                ('log', models.TextField(blank=True)),
                (
                    'batch',
                    models.ForeignKey(
                        blank=True,
                        null=True,
                        on_delete=django.db.models.deletion.CASCADE,
                        to=swapper.get_model_name(
                            'firmware_upgrader', 'BatchUpgradeOperation'
                        ),
                    ),
                ),
                (
                    'device',
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        to=swapper.get_model_name('config', 'Device'),
                    ),
                ),
                (
                    'image',
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        to=get_model_name('FirmwareImage'),
                    ),
                ),
            ],
            options={
                'swappable': swapper.swappable_setting(
                    'firmware_upgrader', 'UpgradeOperation'
                ),
                'abstract': False,
            },
        ),
        migrations.AddField(
            model_name='devicefirmware',
            name='image',
            field=models.ForeignKey(
                on_delete=django.db.models.deletion.CASCADE,
                to=get_model_name('FirmwareImage'),
            ),
        ),
        migrations.AddField(
            model_name='build',
            name='category',
            field=models.ForeignKey(
                help_text='if you have different firmware types eg: (BGP routers, wifi APs, DSL gateways) create a category for each.',
                on_delete=django.db.models.deletion.CASCADE,
                to=get_model_name('Category'),
                verbose_name='firmware category',
            ),
        ),
        migrations.AddField(
            model_name='batchupgradeoperation',
            name='build',
            field=models.ForeignKey(
                on_delete=django.db.models.deletion.CASCADE,
                to=get_model_name('Build'),
            ),
        ),
        migrations.AlterUniqueTogether(
            name='firmwareimage',
            unique_together={('build', 'type')},
        ),
        migrations.AlterUniqueTogether(
            name='category',
            unique_together={('name', 'organization')},
        ),
        migrations.AlterUniqueTogether(
            name='build',
            unique_together={('category', 'version')},
        ),
    ]



from django.urls import include, path

from openwisp_firmware_upgrader import settings as app_settings

urlpatterns = [
    path('', include('openwisp_firmware_upgrader.private_storage.urls')),
]

if app_settings.FIRMWARE_UPGRADER_API:
    urlpatterns += [
        path('api/v1/', include('openwisp_firmware_upgrader.api.urls')),
    ]



from swapper import get_model_name as swapper_get_model_name
from swapper import load_model as swapper_load_model

from .apps import FirmwareUpdaterConfig as AppConfig


def load_model(model):
    return swapper_load_model(AppConfig.label, model)


def get_model_name(model):
    return swapper_get_model_name(AppConfig.label, model)



import uuid

import swapper
from django.contrib.auth import get_user_model
from django.test import Client, TestCase
from django.urls import reverse
from packaging.version import parse as parse_version
from rest_framework import VERSION as REST_FRAMEWORK_VERSION

from openwisp_firmware_upgrader.api.serializers import (
    BatchUpgradeOperationListSerializer,
    BatchUpgradeOperationSerializer,
    BuildSerializer,
    CategorySerializer,
    DeviceFirmwareSerializer,
    DeviceUpgradeOperationSerializer,
    FirmwareImageSerializer,
    UpgradeOperationSerializer,
)
from openwisp_firmware_upgrader.tests.base import TestUpgraderMixin
from openwisp_users.tests.utils import TestMultitenantAdminMixin
from openwisp_utils.tests import AssertNumQueriesSubTestMixin

from ..swapper import load_model

BatchUpgradeOperation = load_model('BatchUpgradeOperation')
Build = load_model('Build')
Category = load_model('Category')
DeviceFirmware = load_model('DeviceFirmware')
FirmwareImage = load_model('FirmwareImage')
UpgradeOperation = load_model('UpgradeOperation')
OrganizationUser = swapper.load_model('openwisp_users', 'OrganizationUser')

user_model = get_user_model()


class TestAPIUpgraderMixin(
    AssertNumQueriesSubTestMixin, TestMultitenantAdminMixin, TestUpgraderMixin
):
    def setUp(self):
        self.org = self._get_org()
        self.operator = self._create_operator(organizations=[self.org])
        self.administrator = self._create_administrator(organizations=[self.org])
        self._login()

    def _obtain_auth_token(self, username='administrator', password='tester'):
        params = {'username': username, 'password': password}
        url = reverse('users:user_auth_token')
        r = self.client.post(url, params)
        self.assertEqual(r.status_code, 200)
        return r.data['token']

    def _login(self, username='administrator', password='tester'):
        token = self._obtain_auth_token(username, password)
        self.client = Client(HTTP_AUTHORIZATION='Bearer ' + token)


class TestBuildViews(TestAPIUpgraderMixin, TestCase):
    def _serialize_build(self, build):
        serializer = BuildSerializer()
        return dict(serializer.to_representation(build))

    def test_build_unauthorized(self):
        build = self._create_build()
        org2 = self._create_org(name='org2', slug='org2')
        OrganizationUser.objects.create(user=self.operator, organization=org2)

        client = Client()
        url = reverse('upgrader:api_build_list')
        with self.subTest(url=url):
            with self.assertNumQueries(0):
                r = client.get(url)
            self.assertEqual(r.status_code, 401)

        url = reverse('upgrader:api_build_detail', args=[build.pk])
        with self.subTest(url=url):
            with self.assertNumQueries(0):
                r = client.get(url)
            self.assertEqual(r.status_code, 401)

    def test_build_list(self):
        self._create_build(organization=self.org)
        self._create_build(version='0.2', organization=self.org)
        serialized_list = [
            self._serialize_build(build)
            for build in Build.objects.all().order_by('-created')
        ]
        url = reverse('upgrader:api_build_list')
        with self.assertNumQueries(5):
            r = self.client.get(url)
        self.assertEqual(r.data['results'], serialized_list)

    def test_build_list_django_filters(self):
        category1 = self._create_category()
        category2 = self._create_category(name='New category')

        build1 = self._create_build(category=category1)
        build2 = self._create_build(version='0.2', category=category2)
        build3 = self._create_build(version='0.2.1', category=category2)
        url = reverse('upgrader:api_build_list')

        filter_params = dict(category=category1.pk)
        with self.assertNumQueries(6):
            r = self.client.get(url, filter_params)
        self.assertEqual(r.data['results'], [self._serialize_build(build1)])

        filter_params = dict(category=category2.pk)
        with self.assertNumQueries(6):
            r = self.client.get(url, filter_params)
        self.assertEqual(
            r.data['results'],
            [self._serialize_build(build3), self._serialize_build(build2)],
        )

        with self.subTest('test version filter'):
            with self.assertNumQueries(5):
                r = self.client.get(url, {'version': '0.2'})
            self.assertEqual(r.data['results'], [self._serialize_build(build2)])

            with self.assertNumQueries(5):
                r = self.client.get(url, {'version': '0.2.1'})
            self.assertEqual(r.data['results'], [self._serialize_build(build3)])

        with self.subTest('test os filter'):
            build1.os = 'abcdefg'
            build1.save(update_fields=('os',))
            build2.os = 'abcdefg-old'
            build2.save(update_fields=('os',))
            with self.assertNumQueries(5):
                r = self.client.get(url, {'os': build1.os})
            self.assertEqual(r.data['results'], [self._serialize_build(build1)])

        with self.subTest('test version, os, category should AND'):
            filter_params.update({'version': '0.2', 'os': 'abcdefg-old'})
            with self.assertNumQueries(6):
                r = self.client.get(url, filter_params)
            self.assertEqual(r.data['results'], [self._serialize_build(build2)])

    def test_build_list_filter_org(self):
        org2 = self._create_org(name='New org', slug='new-org')
        self._create_operator(
            organizations=[org2], username='operator2', email='operator2@test.com'
        )
        cat2 = self._create_category(name='New category', organization=org2)

        build = self._create_build(organization=self.org)
        build2 = self._create_build(version='0.2', category=cat2)

        url = reverse('upgrader:api_build_list')

        self._login('operator', 'tester')
        serialized_list = [
            self._serialize_build(build),
        ]
        with self.assertNumQueries(5):
            r = self.client.get(url)
        self.assertEqual(r.data['results'], serialized_list)

        self._login('operator2', 'tester')
        serialized_list = [
            self._serialize_build(build2),
        ]
        with self.assertNumQueries(5):
            r = self.client.get(url)
        self.assertEqual(r.data['results'], serialized_list)

    def test_build_list_filter_org_admin(self):
        org2 = self._create_org(name='New org', slug='new-org')
        cat2 = self._create_category(name='New category', organization=org2)
        self._create_operator(
            username='admin', email='admin@test.com', is_superuser=True
        )

        self._create_build(organization=self.org)
        build2 = self._create_build(version='0.2', category=cat2)

        url = reverse('upgrader:api_build_list')

        self._login('admin', 'tester')
        serialized_list = [
            self._serialize_build(build)
            for build in Build.objects.all().order_by('-created')
        ]
        with self.assertNumQueries(3):
            r = self.client.get(url)
        self.assertEqual(r.data['results'], serialized_list)

        data_filter = {'organization': 'new-org'}
        serialized_list = [
            self._serialize_build(build2),
        ]
        with self.assertNumQueries(3):
            r = self.client.get(url, data_filter)
        self.assertEqual(r.data['results'], serialized_list)

    def test_build_list_filter_html(self):
        self._create_build(organization=self.org)
        url = reverse('upgrader:api_build_list')
        r = self.client.get(url, HTTP_ACCEPT='text/html')
        # fails if django_filter is not in INSTALLED_APPS
        self.assertEqual(r.status_code, 200)

    def test_build_create(self):
        category = self._get_category()
        url = reverse('upgrader:api_build_list')
        data = {
            'category': category.pk,
            'version': 'asd',
        }
        with self.assertNumQueries(9):
            r = self.client.post(url, data)
        self.assertEqual(Build.objects.count(), 1)
        build = Build.objects.first()
        serialized = self._serialize_build(build)
        self.assertEqual(r.data, serialized)

    def test_build_view(self):
        build = self._create_build()
        serialized = self._serialize_build(build)
        url = reverse('upgrader:api_build_detail', args=[build.pk])
        with self.assertNumQueries(5):
            r = self.client.get(url)
        self.assertEqual(r.data, serialized)

    def test_build_update(self):
        build = self._create_build()
        category = self._get_category()
        url = reverse('upgrader:api_build_detail', args=[build.pk])
        data = {
            'category': str(category.pk),
            'version': '20.04',
            'changelog': 'PUT update',
        }
        with self.assertNumQueries(10):
            r = self.client.put(url, data, content_type='application/json')
        self.assertEqual(r.data['id'], str(build.pk))
        self.assertEqual(r.data['category'], build.category.pk)
        self.assertEqual(r.data['version'], '20.04')
        self.assertEqual(r.data['changelog'], 'PUT update')

    def test_build_update_partial(self):
        build = self._create_build()
        url = reverse('upgrader:api_build_detail', args=[build.pk])
        data = dict(changelog='PATCH update')
        expected_queries = (
            8 if parse_version(REST_FRAMEWORK_VERSION) >= parse_version('3.15') else 9
        )
        with self.assertNumQueries(expected_queries):
            r = self.client.patch(url, data, content_type='application/json')
        self.assertEqual(r.data['id'], str(build.pk))
        self.assertEqual(r.data['category'], build.category.pk)
        self.assertEqual(r.data['version'], build.version)
        self.assertEqual(r.data['changelog'], 'PATCH update')

    def test_build_delete(self):
        build = self._create_build()
        self.assertEqual(Build.objects.count(), 1)
        url = reverse('upgrader:api_build_detail', args=[build.pk])
        r = self.client.delete(url)
        self.assertEqual(r.status_code, 204)
        self.assertEqual(Build.objects.count(), 0)

    def test_api_batch_upgrade(self):
        build = self._create_build()
        self.assertEqual(BatchUpgradeOperation.objects.count(), 0)
        self.assertEqual(DeviceFirmware.objects.count(), 0)
        with self.subTest('Existing build'):
            url = reverse('upgrader:api_build_batch_upgrade', args=[build.pk])
            with self.assertNumQueries(8):
                r = self.client.post(url)
            self.assertEqual(BatchUpgradeOperation.objects.count(), 1)
            self.assertEqual(DeviceFirmware.objects.count(), 0)
            batch = BatchUpgradeOperation.objects.first()
            self.assertEqual(r.status_code, 201)
            self.assertEqual(r.data, {'batch': str(batch.pk)})

        with self.subTest('Non existing build'):
            url = reverse('upgrader:api_build_batch_upgrade', args=[uuid.uuid4()])
            with self.assertNumQueries(4):
                r = self.client.post(url)
            self.assertEqual(r.status_code, 404)

    def test_build_upgradeable(self):
        env = self._create_upgrade_env()
        self.assertEqual(BatchUpgradeOperation.objects.count(), 0)

        url = reverse('upgrader:api_build_batch_upgrade', args=[env['build2'].pk])
        with self.assertNumQueries(10):
            r = self.client.get(url)
        self.assertEqual(r.status_code, 200)
        device_fw_list = [
            str(device_fw.pk)
            for device_fw in DeviceFirmware.objects.all().order_by('-created')
        ]
        self.assertEqual(r.data, {'device_firmwares': device_fw_list, 'devices': []})
        self.assertEqual(BatchUpgradeOperation.objects.count(), 0)

    def test_build_upgradeable_404(self):
        url = reverse('upgrader:api_build_batch_upgrade', args=[uuid.uuid4()])
        with self.assertNumQueries(4):
            r = self.client.get(url)
        self.assertEqual(r.status_code, 404)
        self.assertEqual(BatchUpgradeOperation.objects.count(), 0)


class TestCategoryViews(TestAPIUpgraderMixin, TestCase):
    def _serialize_category(self, category):
        serializer = CategorySerializer()
        return dict(serializer.to_representation(category))

    def test_category_unauthorized(self):
        category = self._create_category()

        org2 = self._create_org(name='org2', slug='org2')
        self.tearDown()
        self.operator.openwisp_users_organization.all().delete()
        OrganizationUser.objects.create(
            user=self.administrator, organization=org2, is_admin=True
        )
        url = reverse('upgrader:api_category_detail', args=[category.pk])
        with self.assertNumQueries(4):
            r = self.client.get(url)
        self.assertEqual(r.status_code, 404)

        client = Client()
        url = reverse('upgrader:api_category_list')
        with self.assertNumQueries(0):
            r = client.get(url)
        self.assertEqual(r.status_code, 401)
        url = reverse('upgrader:api_category_detail', args=[category.pk])
        with self.assertNumQueries(0):
            r = client.get(url)
        self.assertEqual(r.status_code, 401)

    def test_category_list(self):
        self._create_category()
        self._create_category(name='New category')
        serialized_list = [
            self._serialize_category(category)
            for category in Category.objects.all().order_by('-name')
        ]
        url = reverse('upgrader:api_category_list')
        with self.assertNumQueries(5):
            r = self.client.get(url)
        self.assertEqual(r.data['results'], serialized_list)

    def test_category_list_filter_org(self):
        org2 = self._create_org(name='New org', slug='new-org')
        self._create_administrator(
            organizations=[org2],
            username='administrator2',
            email='administrator2@test.com',
        )

        category = self._create_category()
        category2 = self._create_category(name='New category', organization=org2)

        url = reverse('upgrader:api_category_list')

        self._login('administrator', 'tester')
        serialized_list = [
            self._serialize_category(category),
        ]
        with self.assertNumQueries(5):
            r = self.client.get(url)
        self.assertEqual(r.data['results'], serialized_list)

        self._login('administrator2', 'tester')
        serialized_list = [
            self._serialize_category(category2),
        ]
        with self.assertNumQueries(5):
            r = self.client.get(url)
        self.assertEqual(r.data['results'], serialized_list)

    def test_category_list_filter_org_admin(self):
        org2 = self._create_org(name='New org', slug='new-org')
        self._create_operator(
            username='admin', email='admin@test.com', is_superuser=True
        )

        self._create_category()
        category2 = self._create_category(name='New category', organization=org2)

        url = reverse('upgrader:api_category_list')

        self._login('admin', 'tester')
        serialized_list = [
            self._serialize_category(category)
            for category in Category.objects.all().order_by('-name')
        ]
        with self.assertNumQueries(3):
            r = self.client.get(url)
        self.assertEqual(r.data['results'], serialized_list)

        data_filter = {'organization': 'new-org'}
        serialized_list = [
            self._serialize_category(category2),
        ]
        with self.assertNumQueries(3):
            r = self.client.get(url, data_filter)
        self.assertEqual(r.data['results'], serialized_list)

    def test_category_create(self):
        url = reverse('upgrader:api_category_list')
        data = {
            'name': 'Dummy category',
            'organization': self.org.pk,
        }
        with self.assertNumQueries(9):
            r = self.client.post(url, data)
        self.assertEqual(Category.objects.count(), 1)
        category = Category.objects.first()
        serialized = self._serialize_category(category)
        self.assertEqual(r.data, serialized)

    def test_category_view(self):
        category = self._get_category()
        serialized = self._serialize_category(category)
        url = reverse('upgrader:api_category_detail', args=[category.pk])
        with self.assertNumQueries(5):
            r = self.client.get(url)
        self.assertEqual(r.data, serialized)

    def test_category_update(self):
        category = self._get_category()
        url = reverse('upgrader:api_category_detail', args=[category.pk])
        data = {
            'name': 'New name',
            'organization': category.organization.pk,
        }
        with self.assertNumQueries(10):
            r = self.client.put(url, data, content_type='application/json')
        self.assertEqual(r.data['id'], str(category.pk))
        self.assertEqual(r.data['name'], 'New name')
        self.assertEqual(r.data['organization'], category.organization.pk)

    def test_category_update_partial(self):
        category = self._get_category()
        url = reverse('upgrader:api_category_detail', args=[category.pk])
        data = dict(name='New name')
        with self.assertNumQueries(9):
            r = self.client.patch(url, data, content_type='application/json')
        self.assertEqual(r.data['id'], str(category.pk))
        self.assertEqual(r.data['name'], 'New name')
        self.assertEqual(r.data['organization'], category.organization.pk)

    def test_category_delete(self):
        category = self._get_category()
        self.assertEqual(Category.objects.count(), 1)
        url = reverse('upgrader:api_category_detail', args=[category.pk])
        r = self.client.delete(url)
        self.assertEqual(r.status_code, 204)
        self.assertEqual(Category.objects.count(), 0)


class TestBatchUpgradeOperationViews(TestAPIUpgraderMixin, TestCase):
    def _serialize_upgrade_env(self, upgrade_env, action='list'):
        serializer = {
            'list': BatchUpgradeOperationListSerializer,
            'detail': BatchUpgradeOperationSerializer,
        }[action]()
        return dict(serializer.to_representation(upgrade_env))

    def test_batchupgradeoperation_unauthorized(self):
        env = self._create_upgrade_env()
        env['build2'].batch_upgrade(firmwareless=False)

        org2 = self._create_org(name='org2', slug='org2')
        self.tearDown()
        self.administrator.openwisp_users_organization.all().delete()
        OrganizationUser.objects.create(
            user=self.administrator, organization=org2, is_admin=True
        )

        url = reverse(
            'upgrader:api_batchupgradeoperation_detail', args=[env['build2'].pk]
        )
        with self.assertNumQueries(4):
            r = self.client.get(url)
        self.assertEqual(r.status_code, 404)

        client = Client()
        url = reverse('upgrader:api_batchupgradeoperation_list')
        with self.assertNumQueries(0):
            r = client.get(url)
        self.assertEqual(r.status_code, 401)
        url = reverse(
            'upgrader:api_batchupgradeoperation_detail', args=[env['build2'].pk]
        )
        with self.assertNumQueries(0):
            r = client.get(url)
        self.assertEqual(r.status_code, 401)

    def test_batchupgradeoperation_list(self):
        env = self._create_upgrade_env()
        env['build2'].batch_upgrade(firmwareless=False)
        operation = BatchUpgradeOperation.objects.get(build=env['build2'])
        serialized_list = [self._serialize_upgrade_env(operation)]
        url = reverse('upgrader:api_batchupgradeoperation_list')
        with self.assertNumQueries(5):
            r = self.client.get(url)
        self.assertEqual(r.data['results'], serialized_list)

    def test_batchupgradeoperation_list_django_filters(self):
        env = self._create_upgrade_env(organization=self.org)
        env['build1'].batch_upgrade(firmwareless=False)
        env['build2'].batch_upgrade(firmwareless=False)

        url = reverse('upgrader:api_batchupgradeoperation_list')

        serialized_list = [
            self._serialize_upgrade_env(operation)
            for operation in BatchUpgradeOperation.objects.order_by('-created')
        ]
        with self.assertNumQueries(5):
            r = self.client.get(url)
        self.assertEqual(r.data['results'], serialized_list)

        operation = BatchUpgradeOperation.objects.get(build=env['build1'])
        serialized_list = [self._serialize_upgrade_env(operation)]
        filter_params = dict(build=env['build1'].pk)
        with self.assertNumQueries(6):
            r = self.client.get(url, filter_params)
        self.assertEqual(r.data['results'], serialized_list)

        operation = BatchUpgradeOperation.objects.get(build=env['build2'])
        serialized_list = [self._serialize_upgrade_env(operation)]
        filter_params = dict(build=env['build2'].pk)
        with self.assertNumQueries(6):
            r = self.client.get(url, filter_params)
        self.assertEqual(r.data['results'], serialized_list)

        serialized_list = [
            self._serialize_upgrade_env(operation)
            for operation in BatchUpgradeOperation.objects.filter(
                status='in-progress'
            ).order_by('-created')
        ]
        filter_params = dict(status='in-progress')
        with self.assertNumQueries(4):
            r = self.client.get(url, filter_params)
        self.assertEqual(r.data['results'], serialized_list)

        serialized_list = [
            self._serialize_upgrade_env(operation)
            for operation in BatchUpgradeOperation.objects.filter(
                status='success'
            ).order_by('-created')
        ]
        filter_params = dict(status='success')
        with self.assertNumQueries(4):
            r = self.client.get(url, filter_params)
        self.assertEqual(r.data['results'], serialized_list)

    def test_batchupgradeoperation_list_filter_org(self):
        org2 = self._create_org(name='New org', slug='new-org')
        category2 = self._create_category(name='New category', organization=org2)
        self._create_operator(
            organizations=[org2], username='operator2', email='operator2@test.com'
        )

        env = self._create_upgrade_env(organization=self.org)
        env2 = self._create_upgrade_env(category=category2, organization=org2)
        env['build2'].batch_upgrade(firmwareless=False)
        env2['build2'].batch_upgrade(firmwareless=False)

        url = reverse('upgrader:api_batchupgradeoperation_list')

        self._login('operator', 'tester')
        operation = BatchUpgradeOperation.objects.get(build=env['build2'])
        serialized_list = [self._serialize_upgrade_env(operation)]
        with self.assertNumQueries(5):
            r = self.client.get(url)
        self.assertEqual(r.data['results'], serialized_list)

        self._login('operator2', 'tester')
        operation2 = BatchUpgradeOperation.objects.get(build=env2['build2'])
        serialized_list = [self._serialize_upgrade_env(operation2)]
        with self.assertNumQueries(5):
            r = self.client.get(url)
        self.assertEqual(r.data['results'], serialized_list)

    def test_batchupgradeoperation_list_filter_org_admin(self):
        org2 = self._create_org(name='New org', slug='new-org')
        category2 = self._create_category(name='New category', organization=org2)
        self._create_operator(
            username='admin', email='admin@test.com', is_superuser=True
        )

        env = self._create_upgrade_env(organization=self.org)
        env2 = self._create_upgrade_env(category=category2, organization=org2)
        env['build2'].batch_upgrade(firmwareless=False)
        env2['build2'].batch_upgrade(firmwareless=False)

        BatchUpgradeOperation.objects.get(build=env['build2'])
        operation2 = BatchUpgradeOperation.objects.get(build=env2['build2'])

        url = reverse('upgrader:api_batchupgradeoperation_list')

        self._login('admin', 'tester')

        serialized_list = [
            self._serialize_upgrade_env(operation)
            for operation in BatchUpgradeOperation.objects.all().order_by('-created')
        ]
        with self.assertNumQueries(3):
            r = self.client.get(url)
        self.assertEqual(r.data['results'], serialized_list)

        data_filter = {'organization': 'new-org'}
        serialized_list = [self._serialize_upgrade_env(operation2)]
        with self.assertNumQueries(3):
            r = self.client.get(url, data_filter)
        self.assertEqual(r.data['results'], serialized_list)

    def test_batchupgradeoperation_view(self):
        env = self._create_upgrade_env()
        env['build2'].batch_upgrade(firmwareless=False)
        operation = BatchUpgradeOperation.objects.get(build=env['build2'])
        serialized = self._serialize_upgrade_env(operation, action='detail')
        url = reverse('upgrader:api_batchupgradeoperation_detail', args=[operation.pk])
        with self.assertNumQueries(7):
            r = self.client.get(url)
        self.assertEqual(r.data, serialized)


class TestFirmwareImageViews(TestAPIUpgraderMixin, TestCase):
    def _serialize_image(self, firmware):
        serializer = FirmwareImageSerializer()
        data = dict(serializer.to_representation(firmware))
        data['file'] = 'http://testserver' + data['file']
        return data

    def test_firmware_unauthorized(self):
        image = self._create_firmware_image()
        org2 = self._create_org(name='org2', slug='org2')
        OrganizationUser.objects.create(user=self.operator, organization=org2)

        client = Client()
        url = reverse('upgrader:api_firmware_list', args=[image.build.pk])
        with self.subTest(url=url):
            with self.assertNumQueries(1):
                r = client.get(url)
            self.assertEqual(r.status_code, 401)

        url = reverse('upgrader:api_firmware_detail', args=[image.build.pk, image.pk])
        with self.subTest(url=url):
            with self.assertNumQueries(1):
                r = client.get(url)
            self.assertEqual(r.status_code, 401)

        url = reverse('upgrader:api_firmware_download', args=[image.build.pk, image.pk])
        with self.subTest(url=url):
            with self.assertNumQueries(1):
                r = client.get(url)
            self.assertEqual(r.status_code, 401)

    def test_firmware_list(self):
        image = self._create_firmware_image()
        self._create_firmware_image(type=self.TPLINK_4300_IL_IMAGE)

        serialized_list = [
            self._serialize_image(image)
            for image in FirmwareImage.objects.all().order_by('-created')
        ]
        url = reverse('upgrader:api_firmware_list', args=[image.build.pk])
        with self.assertNumQueries(6):
            r = self.client.get(url)
        self.assertEqual(r.data['results'], serialized_list)

    def test_firmware_list_404(self):
        pk = uuid.uuid4()
        url = reverse('upgrader:api_firmware_list', args=[pk])
        r = self.client.get(url)
        with self.subTest(pk=pk):
            self.assertEqual(r.status_code, 404)

    def test_firmware_list_django_filters(self):
        image = self._create_firmware_image(type=self.TPLINK_4300_IMAGE)
        image2 = self._create_firmware_image(type=self.TPLINK_4300_IL_IMAGE)

        url = reverse('upgrader:api_firmware_list', args=[image.build.pk])

        filter_params = dict(type=self.TPLINK_4300_IMAGE)
        with self.assertNumQueries(6):
            r = self.client.get(url, filter_params)
        self.assertEqual(r.data['results'], [self._serialize_image(image)])

        url = reverse('upgrader:api_firmware_list', args=[image.build.pk])

        filter_params = dict(type=self.TPLINK_4300_IL_IMAGE)
        with self.assertNumQueries(6):
            r = self.client.get(url, filter_params)
        self.assertEqual(r.data['results'], [self._serialize_image(image2)])

    def test_firmware_list_filter_org(self):
        org2 = self._create_org(name='New org', slug='new-org')
        self._create_operator(
            organizations=[org2], username='operator2', email='operator2@test.com'
        )
        cat2 = self._create_category(name='New category', organization=org2)

        image = self._create_firmware_image()
        build2 = self._create_build(version='0.2', category=cat2)
        image2 = self._create_firmware_image(build=build2)

        url = reverse('upgrader:api_firmware_list', args=[image.build.pk])

        self._login('operator', 'tester')
        serialized_list = [self._serialize_image(image)]
        with self.assertNumQueries(6):
            r = self.client.get(url)
        self.assertEqual(r.data['results'], serialized_list)

        url = reverse('upgrader:api_firmware_list', args=[image2.build.pk])
        self._login('operator2', 'tester')
        serialized_list = [self._serialize_image(image2)]
        with self.assertNumQueries(6):
            r = self.client.get(url)
        self.assertEqual(r.data['results'], serialized_list)

    def test_firmware_list_filter_org_admin(self):
        org2 = self._create_org(name='New org', slug='new-org')
        self._create_operator(
            username='admin', email='admin@test.com', is_superuser=True
        )
        cat2 = self._create_category(name='New category', organization=org2)

        image = self._create_firmware_image()
        build2 = self._create_build(version='0.2', category=cat2)
        image2 = self._create_firmware_image(build=build2)

        url = reverse('upgrader:api_firmware_list', args=[image.build.pk])

        self._login('admin', 'tester')
        serialized_list = [
            self._serialize_image(image),
        ]

        with self.assertNumQueries(4):
            r = self.client.get(url)
        self.assertEqual(r.data['results'], serialized_list)

        url = reverse('upgrader:api_firmware_list', args=[image2.build.pk])

        data_filter = {'org': 'New org'}
        serialized_list = [self._serialize_image(image2)]
        with self.assertNumQueries(4):
            r = self.client.get(url, data_filter)
        self.assertEqual(r.data['results'], serialized_list)

    def test_firmware_create(self):
        build = self._create_build()
        url = reverse('upgrader:api_firmware_list', args=[build.pk])
        data = {
            # It requires a non-empty file to be uploaded
            'file': self._get_simpleuploadedfile(self.FAKE_IMAGE_PATH2),
            'type': self.TPLINK_4300_IMAGE,
        }
        with self.assertNumQueries(9):
            r = self.client.post(url, data)
        self.assertEqual(r.status_code, 201)
        self.assertEqual(FirmwareImage.objects.count(), 1)
        image = FirmwareImage.objects.first()
        serialized = self._serialize_image(image)
        self.assertEqual(r.data, serialized)

    def test_firmware_create_404(self):
        pk = uuid.uuid4()
        url = reverse('upgrader:api_firmware_list', args=[pk])
        r = self.client.post(
            url,
            {
                'file': self._get_simpleuploadedfile(self.FAKE_IMAGE_PATH2),
                'type': self.TPLINK_4300_IMAGE,
            },
        )
        with self.subTest(pk=pk):
            self.assertEqual(r.status_code, 404)

    def test_firmware_view(self):
        image = self._create_firmware_image()
        serialized = self._serialize_image(image)
        url = reverse('upgrader:api_firmware_detail', args=[image.build.pk, image.pk])
        with self.assertNumQueries(8):
            r = self.client.get(url)
        self.assertEqual(r.data, serialized)

    def test_firmware_delete(self):
        image = self._create_firmware_image()
        self.assertEqual(FirmwareImage.objects.count(), 1)
        url = reverse('upgrader:api_firmware_detail', args=[image.build.pk, image.pk])
        with self.assertNumQueries(11):
            r = self.client.delete(url)
        self.assertEqual(r.status_code, 204)
        self.assertEqual(FirmwareImage.objects.count(), 0)

    def test_firmware_download(self):
        image = self._create_firmware_image()
        with open(self.FAKE_IMAGE_PATH, 'rb') as f:
            content = f.read()
        url = reverse('upgrader:api_firmware_download', args=[image.build.pk, image.pk])
        with self.subTest("Test as operator"):
            self._login('operator', 'tester')
            with self.assertNumQueries(8):
                response = self.client.get(url)
            self.assertEqual(response.getvalue(), content)
        with self.subTest("Test as superuser"):
            self._get_admin()
            self._login('admin', 'tester')
            with self.assertNumQueries(3):
                response = self.client.get(url)
            self.assertEqual(response.getvalue(), content)

    def test_firmware_no_update(self):
        image = self._create_firmware_image()
        url = reverse('upgrader:api_firmware_detail', args=[image.build.pk, image.pk])
        data = {
            'type': self.TPLINK_4300_IL_IMAGE,
            'file': self._get_simpleuploadedfile(),
        }
        r = self.client.put(url, data, content_type='multipart/form-data')
        self.assertEqual(r.status_code, 405)

    def test_firmware_no_update_partial(self):
        image = self._create_firmware_image()
        url = reverse('upgrader:api_firmware_detail', args=[image.build.pk, image.pk])
        data = dict(type=self.TPLINK_4300_IL_IMAGE)
        r = self.client.patch(url, data, content_type='application/json')
        self.assertEqual(r.status_code, 405)


class TestDeviceFirmwareImageViews(TestAPIUpgraderMixin, TestCase):
    def _serialize_device_firmware(self, device_fw):
        serializer = DeviceFirmwareSerializer()
        return dict(serializer.to_representation(device_fw))

    def _create_device_firmware_multi_env(self):
        org1 = self._get_org()
        org2 = self._create_org(name='New org', slug='new-org')
        cat2 = self._create_category(name='New category2', organization=org2)
        build1 = self._get_build()
        build2 = self._create_build(version='0.2', category=cat2)
        image1 = self._create_firmware_image(build=build1)
        image2 = self._create_firmware_image(build=build2)
        d1 = self._create_device(
            name='device1',
            organization=org1,
            mac_address='00:22:bb:33:cc:44',
            model=image1.boards[0],
        )
        d2 = self._create_device(
            name='device2',
            organization=org2,
            mac_address='00:11:bb:22:cc:33',
            model=image2.boards[0],
        )
        ssh_credentials1 = self._get_credentials(organization=org1)
        ssh_credentials2 = self._get_credentials(organization=org2)
        self._create_config(device=d1)
        self._create_config(device=d2)
        self._create_device_connection(device=d1, credentials=ssh_credentials1)
        self._create_device_connection(device=d2, credentials=ssh_credentials2)
        device_fw1 = self._create_device_firmware(
            device=d1, image=image1, device_connection=False
        )
        device_fw2 = self._create_device_firmware(
            device=d2, image=image2, device_connection=False
        )
        self._create_operator(
            organizations=[org1],
            username='org1_manager',
            email='orgmanager@test.com',
        )
        self._create_operator(username='org1_member', email='orgmember@test.com')
        self._create_operator(
            username='org_admin', email='org_admin@test.com', is_superuser=True
        )
        return d1, d2, image1, image2, device_fw1, device_fw2

    def test_device_firmware_detail_unauthorized(self):
        device_fw = self._create_device_firmware()
        client = Client()
        org2 = self._create_org(name='org2', slug='org2')
        OrganizationUser.objects.create(user=self.operator, organization=org2)
        url = reverse('upgrader:api_devicefirmware_detail', args=[device_fw.device.pk])
        with self.subTest(url=url):
            with self.assertNumQueries(0):
                r = client.get(url)
            self.assertEqual(r.status_code, 401)

    def test_device_firmware_detail_404(self):
        device_pk = uuid.uuid4()
        url = reverse('upgrader:api_devicefirmware_detail', args=[device_pk])
        with self.assertNumQueries(4):
            r = self.client.get(url)
        self.assertEqual(r.status_code, 404)

    def test_device_firmware_detail_400(self):
        env = self._create_upgrade_env()
        device1 = env['d1']
        device2 = env['d2']
        image1a = env['image1a']
        device1.model = 'test model'
        device1.full_clean()
        device1.save()

        with self.subTest('Test device and image model validation'):
            url = reverse('upgrader:api_devicefirmware_detail', args=[device1.pk])
            with self.assertNumQueries(18):
                # Try to make a request when the
                # device model does not match the image model
                data = {'image': image1a.pk}
                r = self.client.put(url, data, content_type='application/json')
            self.assertEqual(r.status_code, 400)
            err = 'Device model and image model do not match'
            self.assertIn(err, r.json()['__all__'][0])

        with self.subTest('Test image pk validation'):
            url = reverse('upgrader:api_devicefirmware_detail', args=[device2.pk])
            with self.assertNumQueries(8):
                # image with different "type"
                data = {'image': image1a.pk}
                r = self.client.put(url, data, content_type='application/json')
            self.assertEqual(r.status_code, 400)
            self.assertIn('Invalid pk', r.json()['image'][0])

    def test_deactivated_device(self):
        device_fw = self._create_device_firmware()
        device_fw.device.deactivate()
        url = reverse('upgrader:api_devicefirmware_detail', args=[device_fw.device.pk])

        with self.subTest('Test retrieving DeviceFirmwareImage'):
            response = self.client.get(url)
            self.assertEqual(response.status_code, 200)

        with self.subTest('Test updating DeviceFirmwareImage'):
            response = self.client.put(
                url,
                data={'image': device_fw.image.pk},
                content_type='application/json',
            )
            self.assertEqual(response.status_code, 403)

        with self.subTest('Test deleting DeviceFirmwareImage'):
            response = self.client.delete(url)
            self.assertEqual(response.status_code, 403)

    def test_device_firmware_detail_delete(self):
        device_fw = self._create_device_firmware()
        self.assertEqual(DeviceFirmware.objects.count(), 1)
        url = reverse('upgrader:api_devicefirmware_detail', args=[device_fw.device.pk])
        r = self.client.delete(url)
        self.assertEqual(r.status_code, 204)
        self.assertEqual(DeviceFirmware.objects.count(), 0)

    def test_device_firmware_detail_get(self):
        env = self._create_upgrade_env()
        device1 = env['d1']
        image1a = env['image1a']
        image1b = env['image1b']
        image2a = env['image2a']
        image2b = env['image2b']
        device_fw1 = env['device_fw1']
        category2 = self._get_category(
            organization=self._get_org(), cat_name='Test Category2'
        )
        build2 = self._create_build(category=category2, version='0.2')
        image2 = self._create_firmware_image(build=build2, type=self.TPLINK_4300_IMAGE)

        with self.subTest('Test when device firmware exists'):
            url = reverse(
                'upgrader:api_devicefirmware_detail', args=[device_fw1.device.pk]
            )
            with self.assertNumQueries(9):
                r = self.client.get(url, {'format': 'api'})
            self.assertEqual(r.status_code, 200)
            serializer_detail = self._serialize_device_firmware(device_fw1)
            self.assertEqual(r.data, serializer_detail)
            self.assertContains(r, f'{image1a}</option>')
            self.assertContains(r, f'{image2a}</option>')
            # The "image" field in the browsable API only
            # shows images that are available to the device.
            # This behavior is similar to the "device firmware"
            # inline in the admin interface.
            self.assertNotContains(r, f'{image1b}</option>')
            self.assertNotContains(r, f'{image2b}</option>')
            self.assertNotContains(r, f'{image2}</option>')

        with self.subTest('Test when device firmware does not exist'):
            DeviceFirmware.objects.all().delete()
            url = reverse('upgrader:api_devicefirmware_detail', args=[device1.pk])
            with self.assertNumQueries(8):
                r = self.client.get(url, {'format': 'api'})
            self.assertEqual(r.status_code, 404)
            repsonse = r.content.decode()
            self.assertIn(f'{image1a}</option>', repsonse)
            self.assertIn(f'{image2a}</option>', repsonse)
            self.assertIn(f'{image2}</option>', repsonse)
            # The "image" field in the browsable API only
            # shows images that are available to the device.
            # This behavior is similar to the "device firmware"
            # inline in the admin interface.
            self.assertNotIn(f'{image1b}</option>', repsonse)
            self.assertNotIn(f'{image2b}</option>', repsonse)

    def test_device_firmware_detail_create(self):
        env = self._create_upgrade_env(device_firmware=False)
        device1 = env['d1']
        image1a = env['image1a']
        image1b = env['image1b']
        image2a = env['image2a']
        image2b = env['image2b']
        category2 = self._get_category(
            organization=self._get_org(), cat_name='Test Category2'
        )
        build2 = self._create_build(category=category2, version='0.2')
        image2 = self._create_firmware_image(build=build2, type=self.TPLINK_4300_IMAGE)
        url = reverse('upgrader:api_devicefirmware_detail', args=[device1.pk])
        self.assertEqual(DeviceFirmware.objects.count(), 0)
        self.assertEqual(UpgradeOperation.objects.count(), 0)

        with self.assertNumQueries(26):
            data = {'image': image1a.pk}
            # This API view allows the creation
            # of new devicefirmware objects with
            # a PUT request when the object
            # doesn't already exist.
            r = self.client.put(
                f'{url}?format=api', data, content_type='application/json'
            )

        self.assertEqual(r.status_code, 201)
        self.assertEqual(DeviceFirmware.objects.count(), 1)
        self.assertEqual(UpgradeOperation.objects.count(), 1)
        device_fw1 = DeviceFirmware.objects.first()
        uo1 = UpgradeOperation.objects.first()
        serializer_detail = self._serialize_device_firmware(device_fw1)
        serializer_detail.update({'upgrade_operation': {'id': uo1.id}})
        self.assertEqual(r.data, serializer_detail)
        repsonse = r.content.decode()
        self.assertIn(f'{image1a}</option>', repsonse)
        self.assertIn(f'{image2a}</option>', repsonse)
        self.assertNotIn(f'{image1b}</option>', repsonse)
        self.assertNotIn(f'{image2b}</option>', repsonse)
        self.assertNotIn(f'{image2}</option>', repsonse)

    def test_device_firmware_detail_update(self):
        env = self._create_upgrade_env()
        image1a = env['image1a']
        image1b = env['image1b']
        image2a = env['image2a']
        image2b = env['image2b']
        device_fw1 = env['device_fw1']
        category2 = self._get_category(
            organization=self._get_org(), cat_name='Test Category2'
        )
        build2 = self._create_build(category=category2, version='0.2')
        image2 = self._create_firmware_image(build=build2, type=self.TPLINK_4300_IMAGE)
        url = reverse('upgrader:api_devicefirmware_detail', args=[device_fw1.device.pk])
        self.assertEqual(device_fw1.image.pk, image1a.pk)
        self.assertEqual(DeviceFirmware.objects.count(), 2)
        self.assertEqual(UpgradeOperation.objects.count(), 0)

        with self.assertNumQueries(27):
            data = {'image': image2a.pk}
            r = self.client.put(
                f'{url}?format=api', data, content_type='application/json'
            )

        device_fw1.refresh_from_db()
        self.assertEqual(r.status_code, 200)
        self.assertEqual(device_fw1.image.pk, image2a.pk)
        self.assertEqual(DeviceFirmware.objects.count(), 2)
        self.assertEqual(UpgradeOperation.objects.count(), 1)
        uo1 = UpgradeOperation.objects.first()
        serializer_detail = self._serialize_device_firmware(device_fw1)
        serializer_detail.update({'upgrade_operation': {'id': uo1.id}})
        self.assertEqual(r.data, serializer_detail)
        repsonse = r.content.decode()
        self.assertIn(f'{image1a}</option>', repsonse)
        self.assertIn(f'{image2a}</option>', repsonse)
        self.assertNotIn(f'{image1b}</option>', repsonse)
        self.assertNotIn(f'{image2b}</option>', repsonse)
        self.assertNotIn(f'{image2}</option>', repsonse)

    def test_device_firmware_detail_partial_update(self):
        env = self._create_upgrade_env()
        image1a = env['image1a']
        image1b = env['image1b']
        image2a = env['image2a']
        image2b = env['image2b']
        device_fw1 = env['device_fw1']
        category2 = self._get_category(
            organization=self._get_org(), cat_name='Test Category2'
        )
        build2 = self._create_build(category=category2, version='0.2')
        image2 = self._create_firmware_image(build=build2, type=self.TPLINK_4300_IMAGE)
        url = reverse('upgrader:api_devicefirmware_detail', args=[device_fw1.device.pk])
        self.assertEqual(device_fw1.image.pk, image1a.pk)
        self.assertEqual(DeviceFirmware.objects.count(), 2)
        self.assertEqual(UpgradeOperation.objects.count(), 0)

        with self.assertNumQueries(27):
            data = {'image': image2a.pk}
            r = self.client.patch(
                f'{url}?format=api', data, content_type='application/json'
            )

        device_fw1.refresh_from_db()
        self.assertEqual(r.status_code, 200)
        self.assertEqual(device_fw1.image.pk, image2a.pk)
        self.assertEqual(DeviceFirmware.objects.count(), 2)
        self.assertEqual(UpgradeOperation.objects.count(), 1)
        uo1 = UpgradeOperation.objects.first()
        serializer_detail = self._serialize_device_firmware(device_fw1)
        serializer_detail.update({'upgrade_operation': {'id': uo1.id}})
        self.assertEqual(r.data, serializer_detail)
        repsonse = r.content.decode()
        self.assertIn(f'{image1a}</option>', repsonse)
        self.assertIn(f'{image2a}</option>', repsonse)
        self.assertNotIn(f'{image1b}</option>', repsonse)
        self.assertNotIn(f'{image2b}</option>', repsonse)
        self.assertNotIn(f'{image2}</option>', repsonse)

    def test_device_firmware_detail_multitenancy(self):
        (
            d1,
            d2,
            image1,
            image2,
            device_fw1,
            device_fw2,
        ) = self._create_device_firmware_multi_env()

        with self.subTest('Test device firmware detail org manager'):
            self._login('org1_manager', 'tester')
            url = reverse('upgrader:api_devicefirmware_detail', args=[d1.pk])
            with self.assertNumQueries(7):
                r = self.client.get(url, {'format': 'api'})
            self.assertEqual(r.status_code, 200)
            serializer_detail = self._serialize_device_firmware(device_fw1)
            self.assertEqual(r.data, serializer_detail)
            # The 'org1_manager' user is only
            # authorized to view firmware objects
            self.assertNotContains(r, f'{image1}</option>')
            self.assertNotContains(r, f'{image2}</option>')
            # Only device firmware objects belonging to the
            # same organization as org1_manager can be accessed
            url = reverse('upgrader:api_devicefirmware_detail', args=[d2.pk])
            with self.assertNumQueries(4):
                r = self.client.get(url)
            self.assertEqual(r.status_code, 404)

        with self.subTest('Test device firmware detail org member (403 forbidden)'):
            self._login('org1_member', 'tester')
            url = reverse('upgrader:api_devicefirmware_detail', args=[d1.pk])
            err = 'User is not a manager of the organization'
            with self.assertNumQueries(1):
                r = self.client.get(url)
            self.assertEqual(r.status_code, 403)
            self.assertIn(err, r.json()['detail'])
            url = reverse('upgrader:api_devicefirmware_detail', args=[d2.pk])
            with self.assertNumQueries(1):
                r = self.client.get(url)
            self.assertEqual(r.status_code, 403)
            self.assertIn(err, r.json()['detail'])

        with self.subTest('Test device firmware detail org admin'):
            self._login('org_admin', 'tester')
            url = reverse('upgrader:api_devicefirmware_detail', args=[d1.pk])
            with self.assertNumQueries(6):
                r = self.client.get(url, {'format': 'api'})
            self.assertEqual(r.status_code, 200)
            serializer_detail = self._serialize_device_firmware(device_fw1)
            self.assertEqual(r.data, serializer_detail)
            self.assertContains(r, f'{image1}</option>')
            self.assertNotContains(r, f'{image2}</option>')
            url = reverse('upgrader:api_devicefirmware_detail', args=[d2.pk])
            with self.assertNumQueries(6):
                r = self.client.get(url, {'format': 'api'})
            self.assertEqual(r.status_code, 200)
            serializer_detail = self._serialize_device_firmware(device_fw2)
            self.assertEqual(r.data, serializer_detail)
            self.assertContains(r, f'{image2}</option>')
            self.assertNotContains(r, f'{image1}</option>')


class TestDeviceUpgradeOperationViews(TestAPIUpgraderMixin, TestCase):
    def _serialize_device_upgrade_operation(self, device_uo):
        serializer = DeviceUpgradeOperationSerializer()
        return dict(serializer.to_representation(device_uo))

    def _create_device_uo_multi_env(self):
        org1 = self._get_org()
        org2 = self._create_org(name='New org', slug='new-org')
        cat2 = self._create_category(name='New category2', organization=org2)
        build1 = self._get_build()
        build2 = self._create_build(version='0.2', category=cat2)
        image1 = self._create_firmware_image(build=build1)
        image2 = self._create_firmware_image(build=build2)
        d1 = self._create_device(
            name='device1',
            organization=org1,
            mac_address='00:22:bb:33:cc:44',
            model=image1.boards[0],
        )
        d2 = self._create_device(
            name='device2',
            organization=org2,
            mac_address='00:11:bb:22:cc:33',
            model=image2.boards[0],
        )
        ssh_credentials1 = self._get_credentials(organization=org1)
        ssh_credentials2 = self._get_credentials(organization=org2)
        self._create_config(device=d1)
        self._create_config(device=d2)
        self._create_device_connection(device=d1, credentials=ssh_credentials1)
        self._create_device_connection(device=d2, credentials=ssh_credentials2)
        self._create_device_firmware(
            device=d1, image=image1, upgrade=True, device_connection=False
        )
        self._create_device_firmware(
            device=d2, image=image2, upgrade=True, device_connection=False
        )
        self.assertEqual(UpgradeOperation.objects.count(), 2)
        device_uo1 = UpgradeOperation.objects.get(device=d1)
        device_uo2 = UpgradeOperation.objects.get(device=d2)
        self._create_operator(
            organizations=[org1],
            username='org1_manager',
            email='orgmanager@test.com',
        )
        self._create_operator(username='org1_member', email='orgmember@test.com')
        self._create_operator(
            username='org_admin', email='org_admin@test.com', is_superuser=True
        )
        return d1, d2, device_uo1, device_uo2

    def test_device_uo_list_unauthorized(self):
        device_fw = self._create_device_firmware(upgrade=True)
        client = Client()
        org2 = self._create_org(name='org2', slug='org2')
        OrganizationUser.objects.create(user=self.operator, organization=org2)
        url = reverse(
            'upgrader:api_deviceupgradeoperation_list', args=[device_fw.device.pk]
        )
        with self.subTest(url=url):
            with self.assertNumQueries(1):
                r = client.get(url)
            self.assertEqual(r.status_code, 401)

    def test_device_uo_list_404(self):
        device_pk = uuid.uuid4()
        url = reverse('upgrader:api_deviceupgradeoperation_list', args=[device_pk])
        with self.assertNumQueries(1):
            r = self.client.get(url)
        self.assertEqual(r.status_code, 404)
        self.assertEqual(r.json(), {'detail': 'device not found'})

    def test_device_uo_list_get(self):
        env = self._create_upgrade_env(upgrade_operation=True)
        device1 = env['d1']
        self.assertEqual(UpgradeOperation.objects.count(), 2)
        device_uo1 = UpgradeOperation.objects.get(device=device1)

        with self.subTest('Test when device upgrade operations exist'):
            url = reverse('upgrader:api_deviceupgradeoperation_list', args=[device1.pk])
            with self.assertNumQueries(6):
                r = self.client.get(url)
            self.assertEqual(r.status_code, 200)
            serializer_list = self._serialize_device_upgrade_operation(device_uo1)
            self.assertEqual(r.data['results'], [serializer_list])

        with self.subTest('Test when device upgrade operations does not exist'):
            UpgradeOperation.objects.all().delete()
            url = reverse('upgrader:api_deviceupgradeoperation_list', args=[device1.pk])
            with self.assertNumQueries(5):
                r = self.client.get(url)
            self.assertEqual(r.status_code, 200)
            self.assertEqual(r.data['results'], [])

    def test_device_uo_list_django_filters(self):
        env = self._create_upgrade_env(upgrade_operation=True)
        device1 = env['d1']
        self.assertEqual(UpgradeOperation.objects.count(), 2)
        device_uo1 = UpgradeOperation.objects.get(device=device1)

        with self.subTest('Test filtering using status'):
            url = reverse('upgrader:api_deviceupgradeoperation_list', args=[device1.pk])
            with self.assertNumQueries(6):
                r = self.client.get(url, {'status': 'in-progress'})
            self.assertEqual(r.status_code, 200)
            serializer_list = self._serialize_device_upgrade_operation(device_uo1)
            self.assertEqual(r.data['results'], [serializer_list])
            with self.assertNumQueries(5):
                r = self.client.get(url, {'status': 'failed'})
            self.assertEqual(r.status_code, 200)
            self.assertEqual(r.data['results'], [])

    def test_device_uo_list_multitenancy(self):
        d1, d2, device_uo1, device_uo2 = self._create_device_uo_multi_env()

        with self.subTest('Test device upgrade operation detail org manager'):
            self._login('org1_manager', 'tester')
            url = reverse('upgrader:api_deviceupgradeoperation_list', args=[d1.pk])
            with self.assertNumQueries(6):
                r = self.client.get(url)
            self.assertEqual(r.status_code, 200)
            serializer_list = self._serialize_device_upgrade_operation(device_uo1)
            self.assertEqual(r.data['results'], [serializer_list])
            url = reverse('upgrader:api_deviceupgradeoperation_list', args=[d2.pk])
            with self.assertNumQueries(5):
                r = self.client.get(url)
            self.assertEqual(r.status_code, 200)
            self.assertEqual(r.data['results'], [])

        with self.subTest('Test device upgrade operation org member (403 forbidden)'):
            self._login('org1_member', 'tester')
            url = reverse('upgrader:api_deviceupgradeoperation_list', args=[d1.pk])
            err = 'User is not a manager of the organization'
            with self.assertNumQueries(2):
                r = self.client.get(url)
            self.assertEqual(r.status_code, 403)
            self.assertIn(err, r.json()['detail'])
            url = reverse('upgrader:api_deviceupgradeoperation_list', args=[d2.pk])
            with self.assertNumQueries(2):
                r = self.client.get(url)
            self.assertEqual(r.status_code, 403)
            self.assertIn(err, r.json()['detail'])

        with self.subTest('Test device upgrade operation org admin'):
            self._login('org_admin', 'tester')
            url = reverse('upgrader:api_deviceupgradeoperation_list', args=[d1.pk])
            with self.assertNumQueries(4):
                r = self.client.get(url)
            self.assertEqual(r.status_code, 200)
            serializer_list = self._serialize_device_upgrade_operation(device_uo1)
            self.assertEqual(r.data['results'], [serializer_list])
            url = reverse('upgrader:api_deviceupgradeoperation_list', args=[d2.pk])
            with self.assertNumQueries(4):
                r = self.client.get(url)
            self.assertEqual(r.status_code, 200)
            serializer_list = self._serialize_device_upgrade_operation(device_uo2)
            self.assertEqual(r.data['results'], [serializer_list])


class TestUpgradeOperationViews(TestAPIUpgraderMixin, TestCase):
    def _serialize_upgrade_operation(self, uo, many=False):
        if many:
            serializer = UpgradeOperationSerializer(uo, many=many)
            return serializer.data
        serializer = UpgradeOperationSerializer()
        return dict(serializer.to_representation(uo))

    def _create_upgrade_operation_multi_env(self):
        org1 = self._get_org()
        org2 = self._create_org(name='New org', slug='new-org')
        cat2 = self._create_category(name='New category2', organization=org2)
        build1 = self._get_build()
        build2 = self._create_build(version='0.2', category=cat2)
        image1 = self._create_firmware_image(build=build1)
        image2 = self._create_firmware_image(build=build2)
        d1 = self._create_device(
            name='device1',
            organization=org1,
            mac_address='00:22:bb:33:cc:44',
            model=image1.boards[0],
        )
        d2 = self._create_device(
            name='device2',
            organization=org2,
            mac_address='00:11:bb:22:cc:33',
            model=image2.boards[0],
        )
        ssh_credentials1 = self._get_credentials(organization=org1)
        ssh_credentials2 = self._get_credentials(organization=org2)
        self._create_config(device=d1)
        self._create_config(device=d2)
        self._create_device_connection(device=d1, credentials=ssh_credentials1)
        self._create_device_connection(device=d2, credentials=ssh_credentials2)
        self._create_device_firmware(
            device=d1, image=image1, upgrade=True, device_connection=False
        )
        self._create_device_firmware(
            device=d2, image=image2, upgrade=True, device_connection=False
        )
        self.assertEqual(UpgradeOperation.objects.count(), 2)
        uo1 = UpgradeOperation.objects.get(device=d1)
        uo2 = UpgradeOperation.objects.get(device=d2)
        self._create_operator(
            organizations=[org1],
            username='org1_manager',
            email='orgmanager@test.com',
        )
        self._create_operator(username='org1_member', email='orgmember@test.com')
        self._create_operator(
            username='org_admin', email='org_admin@test.com', is_superuser=True
        )
        return d1, d2, image1, image2, uo1, uo2

    def _assert_uo_list_django_filters(self, query_num, uo, filter_params={}):
        url = reverse('upgrader:api_upgradeoperation_list')
        with self.assertNumQueries(query_num):
            r = self.client.get(url, filter_params)
            self.assertEqual(r.status_code, 200)
            serializer_list = self._serialize_upgrade_operation(uo)
            self.assertEqual(r.data['results'], [serializer_list])

    def test_uo_list_unauthorized(self):
        self._create_device_firmware(upgrade=True)
        client = Client()
        org2 = self._create_org(name='org2', slug='org2')
        OrganizationUser.objects.create(user=self.operator, organization=org2)
        url = reverse('upgrader:api_upgradeoperation_list')
        with self.subTest(url=url):
            with self.assertNumQueries(0):
                r = client.get(url)
            self.assertEqual(r.status_code, 401)

    def test_uo_detail_unauthorized(self):
        device_fw = self._create_device_firmware(upgrade=True)
        client = Client()
        org2 = self._create_org(name='org2', slug='org2')
        OrganizationUser.objects.create(user=self.operator, organization=org2)
        url = reverse(
            'upgrader:api_upgradeoperation_detail', args=[device_fw.device.pk]
        )
        with self.subTest(url=url):
            with self.assertNumQueries(0):
                r = client.get(url)
            self.assertEqual(r.status_code, 401)

    def test_uo_detail_404(self):
        device_pk = uuid.uuid4()
        url = reverse('upgrader:api_upgradeoperation_detail', args=[device_pk])
        with self.assertNumQueries(4):
            r = self.client.get(url)
        self.assertEqual(r.status_code, 404)

    def test_uo_list_get(self):
        self._create_upgrade_env(upgrade_operation=True)
        self.assertEqual(UpgradeOperation.objects.count(), 2)
        uo_qs = UpgradeOperation.objects.order_by('-created')

        with self.subTest('Test when upgrade operations exist'):
            url = reverse('upgrader:api_upgradeoperation_list')
            with self.assertNumQueries(5):
                r = self.client.get(url)
            self.assertEqual(r.status_code, 200)
            serializer_list = self._serialize_upgrade_operation(uo_qs, many=True)
            self.assertEqual(r.data['results'], serializer_list)

        with self.subTest('Test when upgrade operations does not exist'):
            UpgradeOperation.objects.all().delete()
            url = reverse('upgrader:api_upgradeoperation_list')
            with self.assertNumQueries(4):
                r = self.client.get(url)
            self.assertEqual(r.status_code, 200)
            self.assertEqual(r.data['results'], [])

    def test_uo_detail_get(self):
        self._create_upgrade_env(upgrade_operation=True)
        self.assertEqual(UpgradeOperation.objects.count(), 2)
        uo1 = UpgradeOperation.objects.first()

        with self.subTest('Test when upgrade operations exist'):
            url = reverse('upgrader:api_upgradeoperation_detail', args=[uo1.pk])
            with self.assertNumQueries(5):
                r = self.client.get(url)
            self.assertEqual(r.status_code, 200)
            serializer_list = self._serialize_upgrade_operation(uo1)
            self.assertEqual(r.data, serializer_list)

    def test_uo_list_django_filters(self):
        d1, d2, image1, image2, uo1, uo2 = self._create_upgrade_operation_multi_env()
        self.assertEqual(UpgradeOperation.objects.count(), 2)
        self._login('org_admin', 'tester')

        with self.subTest('Test filtering using organization id'):
            self._assert_uo_list_django_filters(
                4, uo1, {'device__organization': d1.organization_id}
            )
            self._assert_uo_list_django_filters(
                4, uo2, {'device__organization': d2.organization_id}
            )

        with self.subTest('Test filtering using organization slug'):
            self._assert_uo_list_django_filters(
                3, uo1, {'device__organization__slug': d1.organization.slug}
            )
            self._assert_uo_list_django_filters(
                3, uo2, {'device__organization__slug': d2.organization.slug}
            )

        with self.subTest('Test filtering using device id'):
            self._assert_uo_list_django_filters(3, uo1, {'device': d1.pk})
            self._assert_uo_list_django_filters(3, uo2, {'device': d2.pk})

        with self.subTest('Test filtering using image id'):
            self._assert_uo_list_django_filters(3, uo1, {'image': image1.pk})
            self._assert_uo_list_django_filters(3, uo2, {'image': image2.pk})

        with self.subTest('Test filtering using status'):
            uo2.status = 'failed'
            uo2.full_clean()
            uo2.save()
            self._assert_uo_list_django_filters(3, uo1, {'status': 'in-progress'})
            self._assert_uo_list_django_filters(3, uo2, {'status': 'failed'})

    def test_uo_list_detail_multitenancy(self):
        _, _, _, _, uo1, uo2 = self._create_upgrade_operation_multi_env()

        with self.subTest('Test upgrade operation list org manager'):
            self._login('org1_manager', 'tester')
            url = reverse('upgrader:api_upgradeoperation_list')
            with self.assertNumQueries(5):
                r = self.client.get(url)
            self.assertEqual(r.status_code, 200)
            serializer_list = self._serialize_upgrade_operation(uo1)
            self.assertEqual(r.data['results'], [serializer_list])

        with self.subTest('Test upgrade operation detail org manager'):
            self._login('org1_manager', 'tester')
            url = reverse('upgrader:api_upgradeoperation_detail', args=[uo1.pk])
            with self.assertNumQueries(5):
                r = self.client.get(url)
            self.assertEqual(r.status_code, 200)
            serializer_detail = self._serialize_upgrade_operation(uo1)
            self.assertEqual(r.data, serializer_detail)
            url = reverse('upgrader:api_upgradeoperation_detail', args=[uo2.pk])
            with self.assertNumQueries(4):
                r = self.client.get(url)
            self.assertEqual(r.status_code, 404)

        with self.subTest('Test upgrade operation list org member (403 forbidden)'):
            self._login('org1_member', 'tester')
            url = reverse('upgrader:api_upgradeoperation_list')
            err = 'User is not a manager of the organization'
            with self.assertNumQueries(1):
                r = self.client.get(url)
            self.assertEqual(r.status_code, 403)
            self.assertIn(err, r.json()['detail'])

        with self.subTest('Test upgrade operation detail org member (403 forbidden)'):
            self._login('org1_member', 'tester')
            url = reverse('upgrader:api_upgradeoperation_detail', args=[uo1.pk])
            err = 'User is not a manager of the organization'
            with self.assertNumQueries(1):
                r = self.client.get(url)
            self.assertEqual(r.status_code, 403)
            self.assertIn(err, r.json()['detail'])
            url = reverse('upgrader:api_upgradeoperation_detail', args=[uo2.pk])
            with self.assertNumQueries(1):
                r = self.client.get(url)
            self.assertEqual(r.status_code, 403)
            self.assertIn(err, r.json()['detail'])

        with self.subTest('Test upgrade operation list org admin'):
            # The org admin can view upgrade
            # operations for both organizations.
            uo_qs = UpgradeOperation.objects.order_by('-created')
            self._login('org_admin', 'tester')
            url = reverse('upgrader:api_upgradeoperation_list')
            with self.assertNumQueries(3):
                r = self.client.get(url)
            self.assertEqual(r.status_code, 200)
            serializer_list = self._serialize_upgrade_operation(uo_qs, many=True)
            self.assertEqual(r.data['results'], serializer_list)
            url = reverse('upgrader:api_upgradeoperation_list')
            with self.assertNumQueries(3):
                r = self.client.get(url)
            self.assertEqual(r.status_code, 200)
            serializer_list = self._serialize_upgrade_operation(uo_qs, many=True)
            self.assertEqual(r.data['results'], serializer_list)


class TestOrgAPIMixin(TestAPIUpgraderMixin, TestCase):
    def _serialize_build(self, build):
        serializer = BuildSerializer()
        return dict(serializer.to_representation(build))

    def test_user_multiple_organizations(self):
        org2 = self._create_org(name='New org', slug='new-org')
        self._create_operator(
            organizations=[self.org, org2],
            username='operator2',
            email='operator2@test.com',
        )

        self._create_build(version='1.0', organization=self.org)
        self._create_build(version='2.0', organization=org2)

        url = reverse('upgrader:api_build_list')

        self._login('operator2', 'tester')
        serialized_list = [
            self._serialize_build(build)
            for build in Build.objects.all().order_by('-created')
        ]
        with self.assertNumQueries(5):
            r = self.client.get(url)
        self.assertEqual(r.data['results'], serialized_list)
        self.assertEqual(r.status_code, 200)


class TestApiMisc(TestAPIUpgraderMixin, TestCase):
    def test_api_docs(self):
        url = reverse('schema-swagger-ui')

        with self.subTest('not authenticated'):
            response = self.client.get(url)
            self.assertEqual(response.status_code, 403)

        with self.subTest('authenticated'):
            self._create_operator(
                username='admin', email='admin@test.com', is_superuser=True
            )
            self._login('admin', 'tester')
            response = self.client.get(url)
            self.assertEqual(response.status_code, 403)



import swapper
from django.test import TestCase
from django.urls import reverse

from openwisp_users.tests.utils import TestMultitenantAdminMixin

from .base import TestUpgraderMixin

OrganizationUser = swapper.load_model('openwisp_users', 'OrganizationUser')


class TestPrivateStorage(TestUpgraderMixin, TestMultitenantAdminMixin, TestCase):
    def setUp(self):
        # Firmware image is created in the default organization
        self.image = self._create_firmware_image()
        self.default_org = self._get_org('default')
        self.test_org = self._get_org()

    def _download_firmware_assert_status(self, status_code):
        response = self.client.get(
            reverse('serve_private_file', args=[self.image.file])
        )
        self.assertEqual(response.status_code, status_code)

    def test_unauthenticated_user(self):
        self._download_firmware_assert_status(403)

    def test_authenticated_user(self):
        user = self._get_user()
        self.client.force_login(user)
        self._download_firmware_assert_status(403)

    def test_authenticated_user_with_different_organization(self):
        self._create_org_user()
        user = self._get_user()
        self.client.force_login(user)
        self._download_firmware_assert_status(403)

    def test_authenticated_user_with_same_organization(self):
        self._create_org_user(organization=self.test_org)
        user = self._get_user()
        self.client.force_login(user)
        self._download_firmware_assert_status(403)

    def test_staff_user_with_different_organization(self):
        staff_user = self._get_operator()
        self._create_org_user(user=staff_user, organization=self.default_org)
        self.client.force_login(staff_user)
        self._download_firmware_assert_status(403)

    def test_staff_user_with_same_organization(self):
        staff_user = self._get_operator()
        self._create_org_user(user=staff_user, organization=self.test_org)
        self.client.force_login(staff_user)
        self._download_firmware_assert_status(403)

    def test_staff_operator_with_different_organization(self):
        staff_user = self._get_operator()
        self._create_org_user(
            user=staff_user, organization=self.default_org, is_admin=True
        )
        self.client.force_login(staff_user)
        self._download_firmware_assert_status(403)

    def test_staff_operator_with_same_organization(self):
        staff_user = self._get_operator()
        self._create_org_user(
            user=staff_user, organization=self.test_org, is_admin=True
        )
        self.client.force_login(staff_user)
        self._download_firmware_assert_status(200)

    def test_superuser(self):
        user = self._get_admin()
        self.client.force_login(user)
        self._download_firmware_assert_status(200)



import io
from contextlib import redirect_stderr, redirect_stdout
from time import sleep
from unittest.mock import patch

from billiard import Queue
from celery.exceptions import Retry
from django.test import TransactionTestCase
from django.utils import timezone
from paramiko.ssh_exception import NoValidConnectionsError, SSHException

from openwisp_controller.connection.connectors.exceptions import CommandFailedException
from openwisp_controller.connection.connectors.openwrt.ssh import (
    OpenWrt as OpenWrtSshConnector,
)
from openwisp_controller.connection.exceptions import NoWorkingDeviceConnectionError
from openwisp_controller.connection.tests.utils import SshServer

from ..swapper import load_model, swapper_load_model
from ..tasks import upgrade_firmware
from ..upgraders.openwrt import OpenWrt
from .base import TestUpgraderMixin, spy_mock

DeviceFirmware = load_model('DeviceFirmware')
DeviceConnection = swapper_load_model('connection', 'DeviceConnection')
Device = swapper_load_model('config', 'Device')


TEST_CHECKSUM = 'e3b0c44298fc1c149afbf4c8996fb92427ae41e4649b934ca495991b7852b855'


def mocked_exec_upgrade_not_needed(command, exit_codes=None):
    cases = {
        f'test -f {OpenWrt.CHECKSUM_FILE}': ['', 0],
        f'cat {OpenWrt.CHECKSUM_FILE}': [TEST_CHECKSUM, 0],
    }
    return cases[command]


def mocked_exec_upgrade_success(command, exit_codes=None, timeout=None):
    defaults = ['', 0]
    _sysupgrade = OpenWrt._SYSUPGRADE
    _checksum = OpenWrt.CHECKSUM_FILE
    cases = {
        'rm -rf /tmp/opkg-lists/': defaults,
        'sync && echo 3 > /proc/sys/vm/drop_caches': defaults,
        'cat /proc/meminfo | grep MemAvailable': ['MemAvailable:      66984 kB', 0],
        f'test -f {_checksum}': defaults,
        f'cat {_checksum}': defaults,
        'mkdir -p /etc/openwisp': defaults,
        f'echo {TEST_CHECKSUM} > {_checksum}': defaults,
        f'{_sysupgrade} --help': ['--test', 1],
        'rm /etc/openwisp/checksum 2> /dev/null': defaults,
        # used in memory check tests
        'test -f /sbin/wifi && /sbin/wifi down': defaults,
        'test -f /sbin/wifi && /sbin/wifi up': defaults,
    }
    if command.startswith(f'{_sysupgrade} --test /tmp/openwrt-'):
        return defaults
    if command.startswith(f'{_sysupgrade} -v -c /tmp/openwrt-'):
        return [
            (
                'Image metadata not found\n'
                'Reading partition table from bootdisk...\n'
                'Reading partition table from image...\n'
            ),
            -1,
        ]
    try:
        return cases[command]
    except KeyError:
        raise CommandFailedException()


def mocked_sysupgrade_failure(command, exit_codes=None, timeout=None):
    if command.startswith(f'{OpenWrt._SYSUPGRADE} -v -c'):
        raise CommandFailedException(
            "Invalid image type\nImage check 'platform_check_image' failed."
        )
    return mocked_exec_upgrade_success(command, exit_codes=None, timeout=None)


def mocked_sysupgrade_test_failure(command, exit_codes=None, timeout=None):
    if command.startswith(f'{OpenWrt._SYSUPGRADE} --test'):
        raise CommandFailedException('Invalid image type')
    return mocked_exec_upgrade_success(command, exit_codes=None, timeout=None)


def mocked_exec_upgrade_memory_success(
    command, exit_codes=None, timeout=None, raise_unexpected_exit=None
):
    global _mock_memory_success_called
    if command.startswith('test -f /etc/init.d/'):
        return ['', 0]
    elif (
        not _mock_memory_success_called
        and command == 'cat /proc/meminfo | grep MemAvailable'
    ):
        _mock_memory_success_called = True
        return ['MemAvailable:      0 kB', 0]
    return mocked_exec_upgrade_success(command, exit_codes=None, timeout=None)


def mocked_exec_upgrade_memory_success_legacy(
    command, exit_codes=None, timeout=None, raise_unexpected_exit=None
):
    global _mock_memory_success_called
    if command == 'cat /proc/meminfo | grep MemAvailable':
        return ['', 1]
    elif command == 'cat /proc/meminfo | grep MemFree':
        if not _mock_memory_success_called:
            _mock_memory_success_called = True
            return ['MemFree:      0 kB', 0]
        else:
            return ['MemFree:      66984 kB', 0]
    return mocked_exec_upgrade_memory_success(
        command, exit_codes, timeout, raise_unexpected_exit
    )


def mocked_exec_upgrade_memory_failure(
    command, exit_codes=None, timeout=None, raise_unexpected_exit=None
):
    if command == 'cat /proc/meminfo | grep MemAvailable':
        return ['MemAvailable:      0 kB', 0]
    return mocked_exec_upgrade_memory_success(command, exit_codes=None, timeout=None)


def mocked_exec_upgrade_memory_aborted(
    command, exit_codes=None, timeout=None, raise_unexpected_exit=None
):
    if command.startswith(f'{OpenWrt._SYSUPGRADE} --test'):
        raise CommandFailedException('Invalid image type')
    return mocked_exec_upgrade_memory_success(command, exit_codes=None, timeout=None)


def mocked_exec_upgrade_success_false_positives(
    command, exit_codes=None, timeout=None, raised_unexpected_exit=None
):
    if command.startswith(f'{OpenWrt._SYSUPGRADE} -v -c /tmp/openwrt-'):
        filename = command.split()[-1].split('/')[-1]
        raise CommandFailedException(
            'Command failed: ubus call system sysupgrade '
            '{ "prefix": "\/tmp\/root", '
            f'"path": "\/tmp\/{filename}", '
            '"backup": "\/tmp\/sysupgrade.tgz", '
            '"command": "\/lib\/upgrade\/do_stage2", '
            '"options": { "save_partitions": 1 } } '
            '(Connection failed)'
        )
    return mocked_exec_upgrade_success(command, exit_codes=None, timeout=None)


def connect_fail_on_write_checksum_pre_action(*args, **kwargs):
    if connect_fail_on_write_checksum.mock.call_count >= 3:
        raise NoValidConnectionsError(errors={'127.0.0.1': 'mocked error'})


_mock_memory_success_called = False
connect_fail_on_write_checksum = spy_mock(
    OpenWrtSshConnector.connect, connect_fail_on_write_checksum_pre_action
)


class TestOpenwrtUpgrader(TestUpgraderMixin, TransactionTestCase):
    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        cls.mock_ssh_server = SshServer(
            {'root': cls._TEST_RSA_PRIVATE_KEY_PATH}
        ).__enter__()
        cls.ssh_server.port = cls.mock_ssh_server.port

    @classmethod
    def tearDownClass(cls):
        cls.mock_ssh_server.__exit__()

    def _trigger_upgrade(self, upgrade=True, exception=None):
        ckey = self._create_credentials_with_key(port=self.ssh_server.port)
        device_conn = self._create_device_connection(credentials=ckey)
        build = self._create_build(organization=device_conn.device.organization)
        image = self._create_firmware_image(build=build)
        output = io.StringIO()
        task_signature = None
        try:
            with redirect_stdout(output):
                device_fw = self._create_device_firmware(
                    image=image,
                    device=device_conn.device,
                    device_connection=False,
                    upgrade=upgrade,
                )
        except Exception as e:
            if exception and isinstance(e, exception):
                device_fw = DeviceFirmware.objects.order_by('created').last()
                if hasattr(e, 'sig'):
                    task_signature = e.sig
            else:
                raise e
        else:
            if exception:
                self.fail(f'{exception.__name__} not raised')

        if not upgrade:
            return device_fw, device_conn, output

        device_conn.refresh_from_db()
        device_fw.refresh_from_db()
        self.assertEqual(device_fw.image.upgradeoperation_set.count(), 1)
        upgrade_op = device_fw.image.upgradeoperation_set.first()
        return device_fw, device_conn, upgrade_op, output, task_signature

    @patch('scp.SCPClient.putfo')
    @patch.object(
        OpenWrt,
        'exec_command',
        side_effect=mocked_sysupgrade_test_failure,
    )
    def test_image_test_failed(self, exec_command, putfo):
        device_fw, device_conn, upgrade_op, output, _ = self._trigger_upgrade()
        self.assertTrue(device_conn.is_working)
        self.assertEqual(exec_command.call_count, 6)
        putfo.assert_called_once()
        self.assertEqual(upgrade_op.status, 'aborted')
        self.assertIn('Invalid image type', upgrade_op.log)
        self.assertFalse(device_fw.installed)

    @patch.object(
        OpenWrt,
        'exec_command',
        side_effect=mocked_exec_upgrade_not_needed,
    )
    def test_upgrade_not_needed(self, mocked):
        device_fw, device_conn, upgrade_op, output, _ = self._trigger_upgrade()
        self.assertTrue(device_conn.is_working)
        self.assertEqual(mocked.call_count, 2)
        self.assertEqual(upgrade_op.status, 'success')
        self.assertIn('upgrade not needed', upgrade_op.log)
        self.assertTrue(device_fw.installed)

    @patch('scp.SCPClient.putfo')
    @patch.object(OpenWrt, 'RECONNECT_DELAY', 0)
    @patch.object(OpenWrt, 'RECONNECT_RETRY_DELAY', 0)
    @patch('billiard.Process.is_alive', return_value=True)
    @patch.object(OpenWrt, 'exec_command', side_effect=mocked_exec_upgrade_success)
    def test_upgrade_success(self, exec_command, is_alive, putfo):
        device_fw, device_conn, upgrade_op, output, _ = self._trigger_upgrade()
        self.assertTrue(device_conn.is_working)
        # should be called 6 times but 1 time is
        # executed in a subprocess and not caught by mock
        self.assertEqual(upgrade_op.status, 'success')
        self.assertEqual(exec_command.call_count, 8)
        self.assertEqual(putfo.call_count, 1)
        self.assertEqual(is_alive.call_count, 1)
        lines = [
            'Image checksum file found',
            'Checksum different, proceeding',
            'Upgrade operation in progress',
            'Trying to reconnect to device at 127.0.0.1 (attempt n.1)',
            'Connected! Writing checksum',
            'Upgrade completed successfully',
        ]
        for line in lines:
            self.assertIn(line, upgrade_op.log)
        self.assertTrue(device_fw.installed)

    @patch.object(OpenWrt, '_call_reflash_command')
    @patch('scp.SCPClient.putfo')
    @patch.object(OpenWrt, 'RECONNECT_DELAY', 0)
    @patch.object(OpenWrt, 'RECONNECT_RETRY_DELAY', 0)
    @patch.object(OpenWrt, 'exec_command', side_effect=mocked_exec_upgrade_success)
    @patch.object(OpenWrtSshConnector, 'connect', connect_fail_on_write_checksum)
    def test_cant_reconnect_on_write_checksum(self, exec_command, putfo, *args):
        start_time = timezone.now()
        with redirect_stderr(io.StringIO()):
            device_fw, device_conn, upgrade_op, output, _ = self._trigger_upgrade()
        self.assertEqual(exec_command.call_count, 6)
        self.assertEqual(putfo.call_count, 1)
        self.assertEqual(connect_fail_on_write_checksum.mock.call_count, 12)
        self.assertEqual(upgrade_op.status, 'failed')
        lines = [
            'Checksum different, proceeding',
            'Upgrade operation in progress',
            'Trying to reconnect to device at 127.0.0.1 (attempt n.1)',
            'Device not reachable yet',
            'Giving up, device not reachable',
        ]
        for line in lines:
            self.assertIn(line, upgrade_op.log)
        self.assertTrue(device_fw.installed)
        self.assertFalse(device_conn.is_working)
        self.assertIn('Giving up', device_conn.failure_reason)
        self.assertTrue(device_conn.last_attempt > start_time)

    @patch('scp.SCPClient.putfo')
    @patch.object(OpenWrt, 'RECONNECT_DELAY', 0)
    @patch.object(OpenWrt, 'RECONNECT_RETRY_DELAY', 0)
    @patch.object(upgrade_firmware, 'max_retries', 1)
    @patch.object(OpenWrt, 'exec_command', side_effect=mocked_exec_upgrade_success)
    @patch.object(
        DeviceConnection,
        'get_working_connection',
        side_effect=NoWorkingDeviceConnectionError(connection=DeviceConnection()),
    )
    def test_connection_failure(self, get_working_connection, exec_command, putfo):
        (
            device_fw,
            device_conn,
            upgrade_op,
            output,
            task_signature,
        ) = self._trigger_upgrade(exception=Retry)
        # retry once for testing purposes
        task_signature.replace().delay()
        upgrade_op.refresh_from_db()
        self.assertFalse(device_conn.is_working)
        self.assertEqual(exec_command.call_count, 0)
        self.assertEqual(putfo.call_count, 0)
        self.assertEqual(get_working_connection.call_count, 2)
        self.assertEqual(upgrade_op.status, 'failed')
        device_conn_error = (
            'Failed to establish connection with the device,'
            ' tried all DeviceConnections.'
        )
        lines = [
            (
                f'Failed to connect with device using {device_conn.credentials}.'
                f' Error: {device_conn.failure_reason}'
            ),
            f'Detected a recoverable failure: {device_conn_error}',
            'The upgrade operation will be retried soon.',
            f'Max retries exceeded. Upgrade failed: {device_conn_error}',
        ]
        for line in lines:
            self.assertIn(line, upgrade_op.log)
        self.assertFalse(device_fw.installed)

    @patch('scp.SCPClient.putfo')
    @patch.object(OpenWrt, 'RECONNECT_DELAY', 0)
    @patch.object(OpenWrt, 'RECONNECT_RETRY_DELAY', 0)
    @patch.object(upgrade_firmware, 'max_retries', 0)
    @patch.object(
        OpenWrtSshConnector,
        'connect',
        side_effect=[
            SSHException('Connection failed'),
            SSHException('Authentication failed'),
        ],
    )
    @patch.object(OpenWrt, 'exec_command', side_effect=mocked_exec_upgrade_success)
    def test_connection_failure_log_all_failure(
        self, mocked_connect, exec_command, putfo
    ):
        org = self._get_org()
        cred1 = self._create_credentials(name='Cred1', organization=org)
        cred2 = self._create_credentials(name='Cred2', organization=org)
        device = self._create_config(organization=org).device
        conn1 = self._create_device_connection(device=device, credentials=cred1)
        conn2 = self._create_device_connection(device=device, credentials=cred2)
        device_fw = self._create_device_firmware(
            device=device,
            device_connection=False,
            upgrade=True,
        )
        upgrade_op = device_fw.image.upgradeoperation_set.first()
        upgrade_op.refresh_from_db()
        lines = [
            (
                f'Failed to connect with device using {conn1.credentials}.'
                f' Error: {conn1.failure_reason}'
            ),
            (
                f'Failed to connect with device using {conn2.credentials}.'
                f' Error: {conn2.failure_reason}'
            ),
            (
                'Max retries exceeded. Upgrade failed:'
                ' Failed to establish connection with the device,'
                ' tried all DeviceConnections.'
            ),
        ]
        for line in lines:
            self.assertIn(line, upgrade_op.log)
        self.assertFalse(device_fw.installed)

    @patch.object(
        OpenWrtSshConnector,
        'upload',
        side_effect=SSHException('Invalid packet blocking'),
    )
    @patch.object(OpenWrt, 'RECONNECT_DELAY', 0)
    @patch.object(OpenWrt, 'RECONNECT_RETRY_DELAY', 0)
    @patch.object(upgrade_firmware, 'max_retries', 1)
    @patch.object(OpenWrt, 'exec_command', side_effect=mocked_exec_upgrade_success)
    def test_upload_failure(self, exec_command, upload):
        (
            device_fw,
            device_conn,
            upgrade_op,
            output,
            task_signature,
        ) = self._trigger_upgrade(exception=Retry)
        task_signature.replace().delay()
        upgrade_op.refresh_from_db()
        self.assertTrue(device_conn.is_working)
        self.assertEqual(upload.call_count, 2)
        self.assertEqual(upgrade_op.status, 'failed')
        lines = [
            'Image checksum file found',
            'Checksum different, proceeding',
            'Detected a recoverable failure: Invalid packet blocking.',
            'The upgrade operation will be retried soon.',
            'Max retries exceeded. Upgrade failed: Invalid packet blocking.',
        ]
        for line in lines:
            self.assertIn(line, upgrade_op.log)
        self.assertFalse(device_fw.installed)

    @patch('openwisp_controller.connection.settings.MANAGEMENT_IP_ONLY', False)
    @patch.object(OpenWrt, '_call_reflash_command')
    @patch('scp.SCPClient.putfo')
    @patch.object(OpenWrt, 'RECONNECT_DELAY', 0)
    @patch.object(OpenWrt, 'RECONNECT_RETRY_DELAY', 0)
    @patch('billiard.Process.is_alive', return_value=True)
    @patch.object(OpenWrt, 'exec_command', side_effect=mocked_exec_upgrade_success)
    def test_device_ip_changed_after_reflash(self, exec_command, alive, putfo, *args):
        device_fw, device_conn, output = self._trigger_upgrade(upgrade=False)

        def connect_pre_action(connector):
            if connect_mocked.mock.call_count == 1:
                return
            # simulate case in which IP address of the device
            # has changed after a few attempts
            if connect_mocked.mock.call_count == 4:
                Device.objects.update(management_ip='192.168.99.254')
            if connect_mocked.mock.call_count > 2:
                raise NoValidConnectionsError(errors={'127.0.0.1': 'mocked error'})

        connect_mocked = spy_mock(OpenWrtSshConnector.connect, connect_pre_action)

        with patch.object(OpenWrtSshConnector, 'connect', connect_mocked):
            with redirect_stderr(io.StringIO()):
                device_fw.save()

        self.assertEqual(device_fw.image.upgradeoperation_set.count(), 1)
        upgrade_op = device_fw.image.upgradeoperation_set.first()
        device_fw.refresh_from_db()

        self.assertEqual(exec_command.call_count, 6)
        self.assertEqual(putfo.call_count, 1)
        self.assertEqual(upgrade_op.status, 'failed')
        lines = [
            'Trying to reconnect to device at 127.0.0.1 (attempt n.1)',
            'Trying to reconnect to device at 127.0.0.1 (attempt n.2)',
            'Trying to reconnect to device at 192.168.99.254, 127.0.0.1 (attempt n.3)',
            'Giving up, device not reachable',
        ]
        for line in lines:
            self.assertIn(line, upgrade_op.log)

    @patch.object(OpenWrt, '_call_reflash_command')
    @patch('scp.SCPClient.putfo')
    @patch('paramiko.SSHClient.connect')
    @patch.object(OpenWrt, 'RECONNECT_DELAY', 0)
    @patch.object(OpenWrt, 'RECONNECT_RETRY_DELAY', 0)
    @patch('billiard.Process.is_alive', return_value=True)
    @patch.object(OpenWrt, 'exec_command', side_effect=mocked_exec_upgrade_success)
    @patch.object(
        DeviceConnection,
        'get_addresses',
        side_effect=[['127.0.0.1'], ['127.0.0.1'], []],
    )
    @patch.object(OpenWrtSshConnector, 'upload')
    def test_device_does_not_have_ip_after_reflash(self, *args):
        _, _, upgrade_op, _, _ = self._trigger_upgrade()
        self.assertNotIn(
            'No valid IP addresses to initiate connections found', upgrade_op.log
        )

    @patch('scp.SCPClient.putfo')
    @patch.object(OpenWrt, 'RECONNECT_DELAY', 0)
    @patch.object(OpenWrt, 'RECONNECT_RETRY_DELAY', 0)
    @patch('billiard.Process.is_alive', return_value=True)
    @patch.object(OpenWrt, 'exec_command', side_effect=mocked_sysupgrade_failure)
    def test_sysupgrade_failure(self, exec_command, is_alive, putfo):
        device_fw, device_conn, upgrade_op, output, _ = self._trigger_upgrade()
        self.assertTrue(device_conn.is_working)
        self.assertEqual(putfo.call_count, 1)
        self.assertEqual(is_alive.call_count, 0)
        self.assertEqual(upgrade_op.status, 'failed')
        lines = [
            'Image checksum file found',
            'Checksum different, proceeding',
            'Upgrade operation in progress',
            'Invalid image type',
            "Image check 'platform_check_image' failed.",
        ]
        for line in lines:
            self.assertIn(line, upgrade_op.log)
        self.assertFalse(device_fw.installed)

    def test_openwrt_settings(self):
        self.assertEqual(OpenWrt.RECONNECT_DELAY, 150)
        self.assertEqual(OpenWrt.RECONNECT_RETRY_DELAY, 30)
        self.assertEqual(OpenWrt.RECONNECT_MAX_RETRIES, 10)
        self.assertEqual(OpenWrt.UPGRADE_TIMEOUT, 80)

    @patch('scp.SCPClient.putfo')
    @patch.object(OpenWrt, 'RECONNECT_DELAY', 0)
    @patch.object(OpenWrt, 'RECONNECT_RETRY_DELAY', 0)
    @patch('billiard.Process.is_alive', return_value=True)
    @patch.object(OpenWrt, 'exec_command', side_effect=mocked_exec_upgrade_success)
    def test_get_upgrade_command(self, exec_command, is_alive, putfo):
        device_fw, device_conn, upgrade_op, output, _ = self._trigger_upgrade()

        with self.subTest('Test upgrade command without upgrade options'):
            upgrade_op.upgrate_options = {}
            upgrader = OpenWrt(upgrade_op, device_conn)
            upgrade_command = upgrader.get_upgrade_command('/tmp/test.bin')
            self.assertEqual(upgrade_command, '/sbin/sysupgrade -v -c /tmp/test.bin')

        with self.subTest('Test upgrade command with upgrade options'):
            upgrade_op.upgrade_options = {
                'c': True,
                'o': False,
                'u': False,
                'n': False,
                'p': False,
                'k': False,
                'F': True,
            }
            upgrader = OpenWrt(upgrade_op, device_conn)
            upgrade_command = upgrader.get_upgrade_command('/tmp/test.bin')
            self.assertEqual(upgrade_command, '/sbin/sysupgrade -v -c -F /tmp/test.bin')

    @patch('scp.SCPClient.putfo')
    @patch.object(OpenWrt, 'RECONNECT_DELAY', 0)
    @patch.object(OpenWrt, 'RECONNECT_RETRY_DELAY', 0)
    @patch('billiard.Process.is_alive', return_value=True)
    def test_call_reflash_command(self, is_alive, putfo):
        with patch.object(
            OpenWrt, 'exec_command', side_effect=mocked_exec_upgrade_success
        ) as exec_command:
            device_fw, device_conn, upgrade_op, output, _ = self._trigger_upgrade()

        upgrader = OpenWrt(upgrade_op, device_conn)
        path = '/tmp/openwrt-image.bin'
        command = f'/sbin/sysupgrade -v -c {path}'

        with self.subTest('success'):
            with patch.object(
                OpenWrt, 'exec_command', side_effect=mocked_exec_upgrade_success
            ) as exec_command:
                failure_queue = Queue()
                OpenWrt._call_reflash_command(
                    upgrader, path, upgrader.UPGRADE_TIMEOUT, failure_queue
                )
                self.assertEqual(exec_command.call_count, 2)
                self.assertEqual(
                    exec_command.call_args_list[0][0],
                    ('rm /etc/openwisp/checksum 2> /dev/null',),
                )
                self.assertEqual(
                    exec_command.call_args_list[0][1], dict(exit_codes=[0, -1, 1])
                )
                self.assertEqual(exec_command.call_args_list[1][0], (command,))
                self.assertEqual(
                    exec_command.call_args_list[1][1],
                    dict(timeout=upgrader.UPGRADE_TIMEOUT, exit_codes=[0, -1]),
                )
                self.assertTrue(failure_queue.empty())
                failure_queue.close()

        with self.subTest('failure'):
            with patch.object(
                OpenWrt, 'exec_command', side_effect=mocked_sysupgrade_failure
            ) as exec_command:
                failure_queue = Queue()
                OpenWrt._call_reflash_command(
                    upgrader, path, upgrader.UPGRADE_TIMEOUT, failure_queue
                )
                self.assertEqual(exec_command.call_count, 2)
                sleep(0.05)
                self.assertFalse(failure_queue.empty())
                exception = failure_queue.get()
                failure_queue.close()
                self.assertIsInstance(exception, CommandFailedException)
                self.assertEqual(
                    str(exception),
                    "Invalid image type\nImage check 'platform_check_image' failed.",
                )

    @patch('scp.SCPClient.putfo')
    @patch.object(OpenWrt, 'RECONNECT_DELAY', 0)
    @patch.object(OpenWrt, 'RECONNECT_RETRY_DELAY', 0)
    @patch('billiard.Process.is_alive', return_value=True)
    @patch.object(
        OpenWrt, 'exec_command', side_effect=mocked_exec_upgrade_memory_success
    )
    def test_upgrade_free_memory_success(self, exec_command, is_alive, putfo):
        global _mock_memory_success_called
        _mock_memory_success_called = False
        device_fw, device_conn, upgrade_op, output, _ = self._trigger_upgrade()
        self.assertTrue(device_conn.is_working)
        self.assertEqual(upgrade_op.status, 'success')
        self.assertEqual(exec_command.call_count, 22)
        self.assertEqual(
            exec_command.call_args_list[5][0][0],
            'test -f /etc/init.d/uhttpd && /etc/init.d/uhttpd stop',
        )
        self.assertEqual(
            exec_command.call_args_list[14][0][0],
            'test -f /etc/init.d/log && /etc/init.d/log stop',
        )
        self.assertEqual(
            exec_command.call_args_list[15][0][0],
            'test -f /sbin/wifi && /sbin/wifi down',
        )
        self.assertEqual(putfo.call_count, 1)
        self.assertEqual(is_alive.call_count, 1)
        lines = [
            'Image checksum file found',
            'Checksum different, proceeding',
            'The image size (0 MiB) is greater than the available memory on the system (0 MiB).',
            'For this reason the upgrade procedure will try to free up',
            'Enough available memory was freed up on the system (65.41 MiB)!',
            'Upgrade operation in progress',
            'Trying to reconnect to device at 127.0.0.1 (attempt n.1)',
            'Connected! Writing checksum',
            'Upgrade completed successfully',
        ]
        for line in lines:
            self.assertIn(line, upgrade_op.log)
        self.assertTrue(device_fw.installed)

    @patch('scp.SCPClient.putfo')
    @patch.object(OpenWrt, 'RECONNECT_DELAY', 0)
    @patch.object(OpenWrt, 'RECONNECT_RETRY_DELAY', 0)
    @patch('billiard.Process.is_alive', return_value=True)
    @patch.object(
        OpenWrt, 'exec_command', side_effect=mocked_exec_upgrade_memory_success_legacy
    )
    def test_upgrade_free_memory_success_legacy(self, exec_command, is_alive, putfo):
        global _mock_memory_success_called
        _mock_memory_success_called = False
        device_fw, device_conn, upgrade_op, output, _ = self._trigger_upgrade()
        self.assertTrue(device_conn.is_working)
        self.assertEqual(upgrade_op.status, 'success')
        self.assertEqual(exec_command.call_count, 24)
        self.assertEqual(
            exec_command.call_args_list[4][0][0],
            'cat /proc/meminfo | grep MemAvailable',
        )
        self.assertEqual(
            exec_command.call_args_list[5][0][0], 'cat /proc/meminfo | grep MemFree'
        )
        self.assertEqual(putfo.call_count, 1)
        self.assertEqual(is_alive.call_count, 1)
        lines = [
            'Image checksum file found',
            'Checksum different, proceeding',
            'The image size (0 MiB) is greater than the available memory on the system (0 MiB).',
            'For this reason the upgrade procedure will try to free up',
            'Enough available memory was freed up on the system (65.41 MiB)!',
            'Upgrade operation in progress',
            'Trying to reconnect to device at 127.0.0.1 (attempt n.1)',
            'Connected! Writing checksum',
            'Upgrade completed successfully',
        ]
        for line in lines:
            self.assertIn(line, upgrade_op.log)
        self.assertTrue(device_fw.installed)

    @patch('scp.SCPClient.putfo')
    @patch.object(OpenWrt, 'RECONNECT_DELAY', 0)
    @patch.object(OpenWrt, 'RECONNECT_RETRY_DELAY', 0)
    @patch('billiard.Process.is_alive', return_value=True)
    @patch.object(
        OpenWrt, 'exec_command', side_effect=mocked_exec_upgrade_memory_failure
    )
    def test_upgrade_free_memory_failure(self, exec_command, is_alive, putfo):
        device_fw, device_conn, upgrade_op, output, _ = self._trigger_upgrade()
        self.assertTrue(device_conn.is_working)
        self.assertEqual(upgrade_op.status, 'aborted')
        self.assertEqual(exec_command.call_count, 30)
        self.assertEqual(
            exec_command.call_args_list[19][0][0],
            'test -f /etc/init.d/uhttpd && /etc/init.d/uhttpd start',
        )
        self.assertEqual(
            exec_command.call_args_list[28][0][0],
            'test -f /etc/init.d/log && /etc/init.d/log start',
        )
        self.assertEqual(
            exec_command.call_args_list[29][0][0],
            'test -f /sbin/wifi && /sbin/wifi up',
        )
        self.assertEqual(putfo.call_count, 0)
        self.assertEqual(is_alive.call_count, 0)
        lines = [
            'Image checksum file found',
            'Checksum different, proceeding',
            'The image size (0 MiB) is greater than the available memory on the system (0 MiB).',
            'For this reason the upgrade procedure will try to free up',
            'There is still not enough available memory on the system (0 MiB)',
            'Starting non critical services again...',
            'Non critical services started, aborting upgrade',
        ]
        for line in lines:
            self.assertIn(line, upgrade_op.log)
        self.assertFalse(device_fw.installed)

    @patch('scp.SCPClient.putfo')
    @patch.object(OpenWrt, 'RECONNECT_DELAY', 0)
    @patch.object(OpenWrt, 'RECONNECT_RETRY_DELAY', 0)
    @patch('billiard.Process.is_alive', return_value=True)
    @patch.object(
        OpenWrt, 'exec_command', side_effect=mocked_exec_upgrade_memory_aborted
    )
    def test_upgrade_free_memory_aborted(self, exec_command, is_alive, putfo):
        global _mock_memory_success_called
        _mock_memory_success_called = False
        device_fw, device_conn, upgrade_op, output, _ = self._trigger_upgrade()
        self.assertTrue(device_conn.is_working)
        self.assertEqual(upgrade_op.status, 'aborted')
        self.assertEqual(exec_command.call_count, 31)
        self.assertEqual(
            exec_command.call_args_list[20][0][0],
            'test -f /etc/init.d/uhttpd && /etc/init.d/uhttpd start',
        )
        self.assertEqual(
            exec_command.call_args_list[29][0][0],
            'test -f /etc/init.d/log && /etc/init.d/log start',
        )
        self.assertEqual(
            exec_command.call_args_list[30][0][0],
            'test -f /sbin/wifi && /sbin/wifi up',
        )
        self.assertEqual(putfo.call_count, 1)
        self.assertEqual(is_alive.call_count, 0)
        lines = [
            'Image checksum file found',
            'Checksum different, proceeding',
            'The image size (0 MiB) is greater than the available memory on the system (0 MiB).',
            'For this reason the upgrade procedure will try to free up',
            'Enough available memory was freed up on the system (65.41 MiB)!',
            'Invalid image type',
            'Starting non critical services again...',
        ]
        for line in lines:
            self.assertIn(line, upgrade_op.log)
        self.assertFalse(device_fw.installed)

    @patch('scp.SCPClient.putfo')
    @patch.object(OpenWrt, 'RECONNECT_DELAY', 0)
    @patch.object(OpenWrt, 'RECONNECT_RETRY_DELAY', 0)
    @patch('billiard.Process.is_alive', return_value=True)
    @patch.object(
        OpenWrt,
        'exec_command',
        side_effect=mocked_exec_upgrade_success_false_positives,
    )
    def test_upgrade_success_false_positives(self, exec_command, is_alive, putfo):
        device_fw, device_conn, upgrade_op, output, _ = self._trigger_upgrade()
        self.assertTrue(device_conn.is_working)
        # should be called 6 times but 1 time is
        # executed in a subprocess and not caught by mock
        self.assertEqual(upgrade_op.status, 'success')
        self.assertEqual(exec_command.call_count, 8)
        self.assertEqual(putfo.call_count, 1)
        self.assertEqual(is_alive.call_count, 1)
        lines = [
            'Image checksum file found',
            'Checksum different, proceeding',
            'Upgrade operation in progress',
            'Trying to reconnect to device at 127.0.0.1 (attempt n.1)',
            'Connected! Writing checksum',
            'Upgrade completed successfully',
        ]
        for line in lines:
            self.assertIn(line, upgrade_op.log)
        self.assertTrue(device_fw.installed)



import json
import logging
from datetime import timedelta

import reversion
import swapper
from django import forms
from django.conf import settings
from django.contrib import admin, messages
from django.contrib.admin.helpers import ACTION_CHECKBOX_NAME
from django.core.serializers.json import DjangoJSONEncoder
from django.shortcuts import redirect
from django.template.response import TemplateResponse
from django.templatetags.static import static
from django.urls import resolve, reverse
from django.utils.html import format_html
from django.utils.safestring import mark_safe
from django.utils.timezone import localtime
from django.utils.translation import get_language
from django.utils.translation import gettext_lazy as _
from reversion.admin import VersionAdmin

from openwisp_controller.config.admin import DeactivatedDeviceReadOnlyMixin, DeviceAdmin
from openwisp_users.multitenancy import MultitenantAdminMixin, MultitenantOrgFilter
from openwisp_utils.admin import ReadOnlyAdmin, TimeReadonlyAdminMixin

from .filters import (
    BuildCategoryFilter,
    BuildCategoryOrganizationFilter,
    CategoryFilter,
    CategoryOrganizationFilter,
)
from .hardware import REVERSE_FIRMWARE_IMAGE_MAP
from .swapper import load_model
from .utils import get_upgrader_schema_for_device
from .widgets import FirmwareSchemaWidget

logger = logging.getLogger(__name__)
BatchUpgradeOperation = load_model('BatchUpgradeOperation')
UpgradeOperation = load_model('UpgradeOperation')
DeviceFirmware = load_model('DeviceFirmware')
FirmwareImage = load_model('FirmwareImage')
Category = load_model('Category')
Build = load_model('Build')
Device = swapper.load_model('config', 'Device')
DeviceConnection = swapper.load_model('connection', 'DeviceConnection')


class BaseAdmin(MultitenantAdminMixin, TimeReadonlyAdminMixin, admin.ModelAdmin):
    save_on_top = True


class BaseVersionAdmin(MultitenantAdminMixin, TimeReadonlyAdminMixin, VersionAdmin):
    history_latest_first = True
    save_on_top = True


@admin.register(load_model('Category'))
class CategoryAdmin(BaseVersionAdmin):
    list_display = ['name', 'organization', 'created', 'modified']
    list_filter = [MultitenantOrgFilter]
    list_select_related = ['organization']
    search_fields = ['name']
    ordering = ['-name', '-created']


class FirmwareImageInline(TimeReadonlyAdminMixin, admin.StackedInline):
    model = FirmwareImage
    extra = 0

    class Media:
        extra = '' if getattr(settings, 'DEBUG', False) else '.min'
        i18n_name = admin.widgets.SELECT2_TRANSLATIONS.get(get_language())
        i18n_file = (
            ('admin/js/vendor/select2/i18n/%s.js' % i18n_name,) if i18n_name else ()
        )
        js = (
            (
                'admin/js/vendor/jquery/jquery%s.js' % extra,
                'admin/js/vendor/select2/select2.full%s.js' % extra,
            )
            + i18n_file
            + ('admin/js/jquery.init.js', 'firmware-upgrader/js/build.js')
        )

        css = {
            'screen': ('admin/css/vendor/select2/select2%s.css' % extra,),
        }

    def has_change_permission(self, request, obj=None):
        if obj:
            return False
        return True


class BatchUpgradeConfirmationForm(forms.ModelForm):
    upgrade_options = forms.JSONField(widget=FirmwareSchemaWidget(), required=False)
    build = forms.ModelChoiceField(
        widget=forms.HiddenInput(), required=False, queryset=Build.objects.all()
    )

    class Meta:
        model = BatchUpgradeOperation
        fields = ('build', 'upgrade_options')

    @property
    def media(self):
        js = [
            'firmware-upgrader/js/upgrade-selected-confirmation.js',
        ]
        css = {'all': ['firmware-upgrader/css/upgrade-selected-confirmation.css']}
        return super().media + forms.Media(js=js, css=css)


@admin.register(load_model('Build'))
class BuildAdmin(BaseAdmin):
    list_display = ['__str__', 'organization', 'category', 'created', 'modified']
    list_filter = [CategoryOrganizationFilter, CategoryFilter]
    list_select_related = ['category', 'category__organization']
    search_fields = ['category__name', 'version', 'os']
    ordering = ['-created', '-version']
    inlines = [FirmwareImageInline]
    actions = ['upgrade_selected']
    multitenant_parent = 'category'
    autocomplete_fields = ['category']

    # Allows apps that extend this modules to use this template with less hacks
    change_form_template = 'admin/firmware_upgrader/change_form.html'

    def organization(self, obj):
        return obj.category.organization

    organization.short_description = _('organization')

    @admin.action(
        description=_('Mass-upgrade devices related to the selected build'),
        permissions=['change'],
    )
    def upgrade_selected(self, request, queryset):
        opts = self.model._meta
        app_label = opts.app_label
        # multiple concurrent batch upgrades are not supported
        # (it's not yet possible to select more builds and upgrade
        #  all of them at the same time)
        if queryset.count() > 1:
            self.message_user(
                request,
                _(
                    'Multiple mass upgrades requested but at the moment only '
                    'a single mass upgrade operation at time is supported.'
                ),
                messages.ERROR,
            )
            # returning None will display the change list page again
            return None
        upgrade_all = request.POST.get('upgrade_all')
        upgrade_related = request.POST.get('upgrade_related')
        upgrade_options = request.POST.get('upgrade_options')
        form = BatchUpgradeConfirmationForm()
        build = queryset.first()
        # upgrade has been confirmed
        if upgrade_all or upgrade_related:
            form = BatchUpgradeConfirmationForm(
                data={'upgrade_options': upgrade_options, 'build': build}
            )
            form.full_clean()
            if not form.errors:
                upgrade_options = form.cleaned_data['upgrade_options']
                batch = build.batch_upgrade(
                    firmwareless=upgrade_all, upgrade_options=upgrade_options
                )
                text = _(
                    'You can track the progress of this mass upgrade operation '
                    'in this page. Refresh the page from time to time to check '
                    'its progress.'
                )
                self.message_user(request, mark_safe(text), messages.SUCCESS)
                url = reverse(
                    f'admin:{app_label}_batchupgradeoperation_change', args=[batch.pk]
                )
                return redirect(url)
        # upgrade needs to be confirmed
        result = BatchUpgradeOperation.dry_run(build=build)
        related_device_fw = result['device_firmwares']
        firmwareless_devices = result['devices']
        title = _('Confirm mass upgrade operation')
        context = self.admin_site.each_context(request)
        upgrader_schema = BatchUpgradeOperation(build=build)._get_upgrader_schema(
            related_device_fw=related_device_fw,
            firmwareless_devices=firmwareless_devices,
        )

        context.update(
            {
                'title': title,
                'related_device_fw': related_device_fw,
                'related_count': len(related_device_fw),
                'firmwareless_devices': firmwareless_devices,
                'firmwareless_count': len(firmwareless_devices),
                'form': form,
                'firmware_upgrader_schema': json.dumps(
                    upgrader_schema, cls=DjangoJSONEncoder
                ),
                'build': build,
                'opts': opts,
                'action_checkbox_name': ACTION_CHECKBOX_NAME,
                'media': self.media,
            }
        )
        request.current_app = self.admin_site.name
        return TemplateResponse(
            request,
            [
                'admin/%s/%s/upgrade_selected_confirmation.html'
                % (app_label, opts.model_name),
                'admin/%s/upgrade_selected_confirmation.html' % app_label,
                'admin/upgrade_selected_confirmation.html',
            ],
            context,
        )

    upgrade_selected.short_description = (
        'Mass-upgrade devices related ' 'to the selected build'
    )

    def change_view(self, request, object_id, form_url='', extra_context=None):
        app_label = self.model._meta.app_label
        extra_context = extra_context or {}
        upgrade_url = f'{app_label}_build_changelist'
        extra_context.update({'upgrade_url': upgrade_url})
        return super().change_view(request, object_id, form_url, extra_context)


class UpgradeOperationForm(forms.ModelForm):
    class Meta:
        fields = ['device', 'image', 'status', 'log', 'modified']
        labels = {'modified': _('last updated')}


class UpgradeOperationInline(admin.StackedInline):
    model = UpgradeOperation
    form = UpgradeOperationForm
    readonly_fields = UpgradeOperationForm.Meta.fields
    extra = 0

    def has_delete_permission(self, request, obj):
        return False

    def has_add_permission(self, request, obj):
        return False

    class Media:
        css = {'all': ['firmware-upgrader/css/upgrade-options.css']}


class ReadonlyUpgradeOptionsMixin:
    @admin.display(description=_('Upgrade options'))
    def readonly_upgrade_options(self, obj):
        upgrader_schema = obj.upgrader_schema
        if not upgrader_schema:
            return _('Upgrade options are not supported for this upgrader.')
        options = []
        for key, value in upgrader_schema['properties'].items():
            option_used = 'yes' if obj.upgrade_options.get(key, False) else 'no'
            option_title = value['title']
            icon_url = static(f'admin/img/icon-{option_used}.svg')
            options.append(
                f'<li><img src="{icon_url}" alt="{option_used}">{option_title}</li>'
            )
        return format_html(
            mark_safe(f'<ul class="readonly-upgrade-options">{"".join(options)}</ul>')
        )


@admin.register(BatchUpgradeOperation)
class BatchUpgradeOperationAdmin(ReadonlyUpgradeOptionsMixin, ReadOnlyAdmin, BaseAdmin):
    list_display = ['build', 'organization', 'status', 'created', 'modified']
    list_filter = [
        BuildCategoryOrganizationFilter,
        'status',
        BuildCategoryFilter,
    ]
    list_select_related = ['build__category__organization']
    ordering = ['-created']
    inlines = [UpgradeOperationInline]
    multitenant_parent = 'build__category'
    fields = [
        'build',
        'status',
        'completed',
        'success_rate',
        'failed_rate',
        'aborted_rate',
        'readonly_upgrade_options',
        'created',
        'modified',
    ]
    autocomplete_fields = ['build']
    readonly_fields = [
        'completed',
        'success_rate',
        'failed_rate',
        'aborted_rate',
        'readonly_upgrade_options',
    ]

    def organization(self, obj):
        return obj.build.category.organization

    organization.short_description = _('organization')

    def get_readonly_fields(self, request, obj):
        fields = super().get_readonly_fields(request, obj)
        return fields + self.__class__.readonly_fields

    def completed(self, obj):
        return obj.progress_report

    def success_rate(self, obj):
        return self.__get_rate(obj.success_rate)

    def failed_rate(self, obj):
        return self.__get_rate(obj.failed_rate)

    def aborted_rate(self, obj):
        return self.__get_rate(obj.aborted_rate)

    def __get_rate(self, value):
        if value:
            return f'{value}%'
        return 'N/A'

    completed.short_description = _('completed')
    success_rate.short_description = _('success rate')
    failed_rate.short_description = _('failure rate')
    aborted_rate.short_description = _('abortion rate')


class DeviceFirmwareForm(forms.ModelForm):
    upgrade_options = forms.JSONField(widget=FirmwareSchemaWidget, required=False)

    class Meta:
        model = DeviceFirmware
        fields = '__all__'

    class Media:
        js = ['admin/js/jquery.init.js', 'firmware-upgrader/js/device-firmware.js']
        css = {'all': ['firmware-upgrader/css/device-firmware.css']}

    def _get_image_queryset(self, device):
        # restrict firmware images to organization of the current device
        qs = (
            FirmwareImage.objects.filter(
                build__category__organization_id=device.organization_id
            )
            .order_by('-created')
            .select_related('build', 'build__category')
        )
        # if device model is defined
        # restrict the images to the ones compatible with it
        if device.model and device.model in REVERSE_FIRMWARE_IMAGE_MAP:
            qs = qs.filter(type=REVERSE_FIRMWARE_IMAGE_MAP[device.model])
        # if DeviceFirmware instance already exists
        # restrict images to the ones of the same category
        if not self.instance._state.adding:
            self.instance.refresh_from_db()
            qs = qs.filter(build__category_id=self.instance.image.build.category_id)
        return qs

    def __init__(self, device, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.fields['image'].queryset = self._get_image_queryset(device)

    def full_clean(self):
        super().full_clean()
        if not self.errors and hasattr(self, 'cleaned_data'):
            upgrade_op = UpgradeOperation(
                device=self.cleaned_data['device'],
                image=self.cleaned_data['image'],
                upgrade_options=self.cleaned_data['upgrade_options'],
            )
            try:
                upgrade_op.full_clean()
            except forms.ValidationError as error:
                self.add_error('__all__', error.messages[0])

    def save(self, commit=True):
        """
        Adapted from ModelForm.save()
        Passes "upgrade_options to DeviceFirmware.save()
        """
        if commit:
            # If committing, save the instance and the m2m data immediately.
            self.instance.save(upgrade_options=self.cleaned_data['upgrade_options'])
        return self.instance


class DeviceFormSet(forms.BaseInlineFormSet):
    def get_form_kwargs(self, index):
        kwargs = super().get_form_kwargs(index)
        kwargs['device'] = self.instance
        return kwargs


class DeviceFirmwareInline(
    MultitenantAdminMixin, DeactivatedDeviceReadOnlyMixin, admin.StackedInline
):
    model = DeviceFirmware
    formset = DeviceFormSet
    form = DeviceFirmwareForm
    exclude = ['created']
    select_related = ['device', 'image']
    readonly_fields = ['installed', 'modified']
    verbose_name = _('Firmware')
    verbose_name_plural = verbose_name
    extra = 0
    multitenant_shared_relations = ['device']
    template = 'admin/firmware_upgrader/device_firmware_inline.html'
    # hack for openwisp-monitoring integartion
    # TODO: remove when this issue solved:
    # https://github.com/theatlantic/django-nested-admin/issues/128#issuecomment-665833142
    sortable_options = {'disabled': True}

    def _get_conditional_queryset(self, request, obj, select_related=False):
        return bool(obj)

    def get_formset(self, request, obj=None, **kwargs):
        formset = super().get_formset(request, obj=obj, **kwargs)
        if obj:
            try:
                schema = get_upgrader_schema_for_device(obj)
                formset.extra_context = json.dumps(schema, cls=DjangoJSONEncoder)
            except DeviceConnection.DoesNotExist:
                # We cannot retrieve the schema for upgrade options because this
                # device does not have any related DeviceConnection object.
                pass
        return formset


class DeviceUpgradeOperationForm(UpgradeOperationForm):
    class Meta(UpgradeOperationForm.Meta):
        pass

    def __init__(self, device, *args, **kwargs):
        self.device = device
        super().__init__(*args, **kwargs)


class DeviceUpgradeOperationInline(ReadonlyUpgradeOptionsMixin, UpgradeOperationInline):
    verbose_name = _('Recent Firmware Upgrades')
    verbose_name_plural = verbose_name
    formset = DeviceFormSet
    form = DeviceUpgradeOperationForm
    # hack for openwisp-monitoring integration
    # TODO: remove when this issue solved:
    # https://github.com/theatlantic/django-nested-admin/issues/128#issuecomment-665833142
    sortable_options = {'disabled': True}
    fields = [
        'device',
        'image',
        'status',
        'log',
        'readonly_upgrade_options',
        'modified',
    ]
    readonly_fields = fields

    def get_queryset(self, request, select_related=True):
        """
        Return recent upgrade operations for this device
        (created within the last 7 days)
        """
        qs = super().get_queryset(request)
        resolved = resolve(request.path_info)
        if 'object_id' in resolved.kwargs:
            seven_days = localtime() - timedelta(days=7)
            qs = qs.filter(
                device_id=resolved.kwargs['object_id'], created__gte=seven_days
            ).order_by('-created')
        if select_related:
            qs = qs.select_related()
        return qs

    def _get_conditional_queryset(self, request, obj, select_related=False):
        if obj:
            return self.get_queryset(request, select_related=False).exists()
        return False


# DeviceAdmin.get_inlines = device_admin_get_inlines
DeviceAdmin.conditional_inlines += [DeviceFirmwareInline, DeviceUpgradeOperationInline]

reversion.register(model=DeviceFirmware, follow=['device'])
reversion.register(model=UpgradeOperation)
DeviceAdmin.add_reversion_following(follow=['devicefirmware', 'upgradeoperation_set'])



# Generated by Django 3.0.5 on 2020-06-02 17:56

from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ('firmware_upgrader', '0003_private_media'),
    ]

    operations = [
        migrations.AlterField(
            model_name='batchupgradeoperation',
            name='status',
            field=models.CharField(
                choices=[
                    ('idle', 'idle'),
                    ('in-progress', 'in progress'),
                    ('success', 'completed successfully'),
                    ('failed', 'completed with some failures'),
                ],
                default='idle',
                max_length=12,
            ),
        ),
    ]



from swapper import swappable_setting

from .base.models import (
    AbstractBatchUpgradeOperation,
    AbstractBuild,
    AbstractCategory,
    AbstractDeviceFirmware,
    AbstractFirmwareImage,
    AbstractUpgradeOperation,
)


class Category(AbstractCategory):
    class Meta(AbstractCategory.Meta):
        abstract = False
        swappable = swappable_setting('firmware_upgrader', 'Category')


class Build(AbstractBuild):
    class Meta(AbstractBuild.Meta):
        abstract = False
        swappable = swappable_setting('firmware_upgrader', 'Build')


class FirmwareImage(AbstractFirmwareImage):
    class Meta(AbstractFirmwareImage.Meta):
        abstract = False
        swappable = swappable_setting('firmware_upgrader', 'FirmwareImage')


class DeviceFirmware(AbstractDeviceFirmware):
    class Meta(AbstractDeviceFirmware.Meta):
        abstract = False
        swappable = swappable_setting('firmware_upgrader', 'DeviceFirmware')


class BatchUpgradeOperation(AbstractBatchUpgradeOperation):
    class Meta(AbstractBatchUpgradeOperation.Meta):
        abstract = False
        swappable = swappable_setting('firmware_upgrader', 'BatchUpgradeOperation')


class UpgradeOperation(AbstractUpgradeOperation):
    class Meta(AbstractUpgradeOperation.Meta):
        abstract = False
        swappable = swappable_setting('firmware_upgrader', 'UpgradeOperation')



from django.conf import settings
from django.core.exceptions import ImproperlyConfigured
from django.utils.module_loading import import_string

from openwisp_controller.connection import settings as conn_settings

CUSTOM_OPENWRT_IMAGES = getattr(settings, 'OPENWISP_CUSTOM_OPENWRT_IMAGES', None)
# fmt: off
UPGRADERS_MAP = getattr(settings, 'OPENWISP_FIRMWARE_UPGRADERS_MAP', {
    conn_settings.DEFAULT_UPDATE_STRATEGIES[0][0]: 'openwisp_firmware_upgrader.upgraders.openwrt.OpenWrt',
    conn_settings.DEFAULT_UPDATE_STRATEGIES[1][0]: 'openwisp_firmware_upgrader.upgraders.openwisp.OpenWisp1'
})
# fmt: on

MAX_FILE_SIZE = getattr(
    settings, 'OPENWISP_FIRMWARE_UPGRADER_MAX_FILE_SIZE', 30 * 1024 * 1024
)

RETRY_OPTIONS = getattr(
    settings,
    'OPENWISP_FIRMWARE_UPGRADER_RETRY_OPTIONS',
    dict(max_retries=4, retry_backoff=60, retry_backoff_max=600, retry_jitter=True),
)

TASK_TIMEOUT = getattr(settings, 'OPENWISP_FIRMWARE_UPGRADER_TASK_TIMEOUT', 1500)

FIRMWARE_UPGRADER_API = getattr(settings, 'OPENWISP_FIRMWARE_UPGRADER_API', True)
FIRMWARE_API_BASEURL = getattr(settings, 'OPENWISP_FIRMWARE_API_BASEURL', '/')
OPENWRT_SETTINGS = getattr(settings, 'OPENWISP_FIRMWARE_UPGRADER_OPENWRT_SETTINGS', {})

# Path of urls that need to be refered in migrations files.
IMAGE_URL_PATH = 'firmware/'

try:
    PRIVATE_STORAGE_INSTANCE = import_string(
        getattr(
            settings,
            'OPENWISP_FIRMWARE_PRIVATE_STORAGE_INSTANCE',
            'openwisp_firmware_upgrader.private_storage.storage.file_system_private_storage',
        )
    )
except ImportError:
    raise ImproperlyConfigured(
        'Failed to import FIRMWARE_UPGRADER_PRIVATE_STORAGE_INSTANCE'
    )



import shlex
import subprocess

from django.utils.translation import gettext_lazy as _

from .openwrt import OpenWrt


class OpenWisp1(OpenWrt):
    """
    Upgrader for OpenWISP 1.x
    Used to migrate legacy OpenWISP 1.x systems
    (the previous generation of OpenWISP built in ruby on rails)
    to OpenWISP 2.
    """

    UPGRADE_COMMAND = '{sysupgrade} -c -n {path}'
    RECONNECT_DELAY = 60 * 6

    def _test_image(self, path):  # pragma: no cover
        # ensure sysupgrade --test is supported or skip
        help_text, code = self.exec_command(
            f'{self._SYSUPGRADE} --help', exit_codes=[1]
        )
        if 'Usage:' in help_text and '--test' not in help_text:
            self.log(
                _(
                    'This image does not support sysupgrade --test, skipping this step...'
                )
            )
        else:
            super()._test_image(path)

    def _reflash_legacy(self, path, timeout):  # pragma: no cover
        self.log(
            _(
                'The version used is OpenWrt Backfire, '
                'using legacy reflash instructions.'
            )
        )

        credentials = self.connection.credentials.params
        if 'key' not in credentials:
            raise ValueError('SSH Key not found in credentials')
        ssh_private_key = credentials['key'] + '\n'

        # create temporary file
        process = subprocess.Popen('mktemp', stdout=subprocess.PIPE)
        result = process.communicate(timeout=5)
        if process.returncode != 0:
            raise ValueError(f'mktemp exited with {process.returncode}')

        # write SSH key to the temporary file
        temp_file_path = result[0].decode().strip()
        with open(temp_file_path, 'w') as temp_file:
            temp_file.write(ssh_private_key)

        # get sysupgrade command text
        sysupgrade = self.get_upgrade_command(path)
        # remove -c because not supported on backfire
        sysupgrade = sysupgrade.replace('-c ', '')
        # $PATH is buggy on Backfire,
        # a shell command in a subprocess
        # that sets the right path fixes it
        # without this, the upgrade fails
        ip = self.addresses[0]
        path = '/bin:/sbin:/usr/bin:/usr/sbin'
        command = (
            'ssh -o StrictHostKeyChecking=no '
            '-o UserKnownHostsFile=/dev/null '
            '-o "IdentitiesOnly=yes" '
            f'-i {temp_file_path} '
            f'root@{ip} -T '
            f'"export PATH={path}; {sysupgrade}"'
        )
        args = shlex.split(command)
        process = subprocess.Popen(
            args=args, stderr=subprocess.PIPE, stdout=subprocess.PIPE
        )
        output_results = list(process.communicate(timeout=timeout))

        # delete tmp file
        subprocess.Popen(shlex.split(f'rm {temp_file_path}'))

        # if there's any error, raise an exception
        output_results.reverse()
        output = ''
        for output_result in output_results:
            output += output_result.decode()
        if process.returncode != 0:
            raise ValueError(output)
        # log output if there was no error
        self.log(output)

    @classmethod
    def _call_reflash_command(
        cls, upgrader, path, timeout, failure_queue
    ):  # pragma: no cover
        upgrader.connect()
        # ensure these files are preserved after the upgrade
        upgrader.exec_command('echo /etc/config/network >> /etc/sysupgrade.conf')
        upgrader.exec_command(
            'echo /etc/dropbear/dropbear_rsa_host_key >> /etc/sysupgrade.conf'
        )
        upgrader.log(
            _(
                'Written openwisp config file in /etc/config/openwisp.\n'
                'Added entries to /etc/sysupgrade.conf:\n'
                '- /etc/config/network\n'
                '- /etc/dropbear/dropbear_rsa_host_key\n'
            )
        )
        output, exit_code = upgrader.exec_command(
            'cat /etc/openwrt_release', raise_unexpected_exit=False
        )
        try:
            if output and 'backfire' in output:
                upgrader._reflash_legacy(path, timeout=timeout)
            else:
                command = upgrader.get_upgrade_command(path)
                output, exit_code = upgrader.exec_command(
                    command, timeout=timeout, exit_codes=[0, -1]
                )
                upgrader.log(output)
        except Exception as e:
            failure_queue.put(e)
        upgrader.disconnect()
